#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Post-simulation single-cell plot pipelines** (i.e., pipelines plotting
simulated data of a single cell of the cell cluster).
'''

# ....................{ IMPORTS                           }....................
import numpy as np
from betse.exceptions import BetseSimConfException
from betse.science.config.export.visual.confexpvisplot import (
    SimConfExportPlotCell)
from betse.science.phase.phasecls import SimPhase
from betse.science.phase.require import phasereqs
from betse.science.pipe.export.plot.pipeexpplotabc import SimPipeExportPlotABC
from betse.science.pipe.piperun import piperunner
from betse.science.visual.plot import plotutil
from betse.util.type.descriptor.descs import classproperty_readonly
from betse.util.type.types import type_check, SequenceTypes
from matplotlib import pyplot

# ....................{ SUBCLASSES                        }....................
class SimPipeExportPlotCell(SimPipeExportPlotABC):
    '''
    **Post-simulation single-cell plot pipeline** (i.e., object iteratively
    displaying and/or saving all plots specific to a single cell of the cell
    cluster, produced after initialization and simulation as specified by the
    current simulation configuration).
    '''

    # ..................{ SUPERCLASS ~ properties           }..................
    @classproperty_readonly
    def _NOUN_SINGULAR(cls) -> str:
        return 'single-cell plot'

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def init(self, phase: SimPhase) -> None:

        # Initialize our superclass for the current call to the _run() method.
        super().init(phase)

        # If this index is not that of an actual cell, raise an exception.
        if phase.p.visual.single_cell_index not in phase.cells.cell_i:
            raise BetseSimConfException(
                'Plot cell index {} invalid '
                '(i.e., not in range [{}, {}]).'.format(
                    phase.p.visual.single_cell_index,
                    phase.cells.cell_i[0],
                    phase.cells.cell_i[-1]))

    # ..................{ SUPERCLASS                        }..................
    @type_check
    def iter_runners_conf(self, phase: SimPhase) -> SequenceTypes:
        return phase.p.plot.plots_cell_after_sim

    # ..................{ EXPORTERS ~ cell : current        }..................
    @piperunner(
        categories=('Current Density', 'Transmembrane',),
        requirements=phasereqs.ELECTRIC_CURRENT_MEMBRANE,
    )
    def export_currents_membrane(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot all transmembrane current densities for the single cell indexed by
        the current simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        pyplot.figure()
        axI = pyplot.subplot(111)

        if phase.p.is_ecm:
            # Total cell current storage vector.
            Imem = []

            # Indices of all membranes for this cell.
            mems_for_plotcell = phase.cells.cell_to_mems[
                phase.p.visual.single_cell_index]

            for t in range(len(phase.sim.time)):
                memArray = phase.sim.I_mem_time[t]

                # X and Y components of the net current at each membrane.
                Ixo = (
                    memArray[mems_for_plotcell] *
                    phase.cells.mem_vects_flat[mems_for_plotcell,2] *
                    phase.cells.mem_sa[mems_for_plotcell]
                )
                Iyo = (
                    memArray[mems_for_plotcell] *
                    phase.cells.mem_vects_flat[mems_for_plotcell,3] *
                    phase.cells.mem_sa[mems_for_plotcell]
                )

                # X and Y components of the net current over all membranes,
                # implicitly accounting for in-out directionality.
                Ix = np.sum(Ixo)
                Iy = np.sum(Iyo)

                # Current density, obtained by dividing the total magnitude of
                # the net current by this cell's surface area.
                Io = np.sqrt(Ix**2 + Iy**2) / phase.cells.cell_sa[
                    phase.p.visual.single_cell_index]
                Imem.append(100*Io)
        else:
            Imem = [
                100*memArray[phase.p.visual.single_cell_index]
                for memArray in phase.sim.I_mem_time
            ]

        axI.plot(phase.sim.time, Imem)
        axI.set_xlabel('Time [s]')
        axI.set_ylabel('Current density [uA/cm2]')
        axI.set_title(
            'Transmembrane current density for cell {}'.format(
                phase.p.visual.single_cell_index))

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='Imem_time')

    # ..................{ EXPORTERS ~ cell : deform         }..................
    @piperunner(
        categories=('Deformation', 'Total',),
        requirements=phasereqs.DEFORM,
    )
    def export_deform_total(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot all **total cellular displacements** (i.e., summations of all
        cellular deformations due to galvanotropic and osmotic pressure body
        forces) for the single cell indexed by the current simulation
        configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        # Extract time-series deformation data for the plot cell.
        dx = np.asarray([
            arr[phase.p.visual.single_cell_index]
            for arr in phase.sim.dx_cell_time])
        dy = np.asarray([
            arr[phase.p.visual.single_cell_index]
            for arr in phase.sim.dy_cell_time])

        # Get the total magnitude.
        disp = np.sqrt(dx**2 + dy**2)

        pyplot.figure()
        axD = pyplot.subplot(111)
        axD.plot(phase.sim.time, phase.p.um*disp)
        axD.set_xlabel('Time [s]')
        axD.set_ylabel('Displacement [um]')
        axD.set_title('Displacement of cell {}'.format(
            phase.p.visual.single_cell_index))

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='Displacement_time')

    # ..................{ EXPORTERS ~ cell : ion            }..................
    @piperunner(
        categories=('Ion Concentration', 'M anion'),
        requirements=phasereqs.ION_M_ANION,
    )
    def export_ion_m_anion(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot all M anion (i.e., M-) ion concentrations for the single cell
        indexed by the current simulation configuration over all sampled time
        steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        figConcsM, axConcsM = plotutil.plotSingleCellCData(
            phase.sim.cc_time,
            phase.sim.time,
            phase.sim.iM,
            phase.p.visual.single_cell_index,
            fig=None, ax=None,
            lncolor='r',
            ionname='M-',
        )
        axConcsM.set_title(
            'M Anion concentration in cell {}'.format(phase.p.visual.single_cell_index))

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='concM_time')


    @piperunner(
        categories=('Ion Concentration', 'Potassium'),
        requirements=phasereqs.ION_POTASSIUM,
    )
    def export_ion_potassium(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot all potassium (i.e., K+) ion concentrations for the single cell
        indexed by the current simulation configuration over all sampled time
        steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        figConcsK, axConcsK = plotutil.plotSingleCellCData(
            phase.sim.cc_time,
            phase.sim.time,
            phase.sim.iK,
            phase.p.visual.single_cell_index,
            fig=None, ax=None,
            lncolor='b',
            ionname='K+',
        )
        axConcsK.set_title(
            'Potassium concentration in cell {}'.format(
                phase.p.visual.single_cell_index))

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='concK_time')


    @piperunner(
        categories=('Ion Concentration', 'Sodium'),
        requirements=phasereqs.ION_SODIUM,
    )
    def export_ion_sodium(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot all sodium (i.e., Na+) ion concentrations for the single cell
        indexed by the current simulation configuration over all sampled time
        steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        # Plot cell sodium concentration versus time.
        figConcsNa, axConcsNa = plotutil.plotSingleCellCData(
            phase.sim.cc_time,
            phase.sim.time,
            phase.sim.iNa,
            phase.p.visual.single_cell_index,
            fig=None,
            ax=None,
            lncolor='g',
            ionname='Na+',
        )
        axConcsNa.set_title(
            'Sodium concentration in cell {}'.format(phase.p.visual.single_cell_index))

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='concNa_time')

    # ..................{ EXPORTERS ~ cell : ion : calcium  }..................
    @piperunner(
        categories=('Ion Concentration', 'Calcium', 'Cellular',),
        requirements=phasereqs.ION_CALCIUM,
    )
    def export_ion_calcium(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot all calcium (i.e., Ca2+) ion concentrations for the single cell
        indexed by the current simulation configuration over all sampled time
        steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        figA, axA = plotutil.plotSingleCellCData(
            phase.sim.cc_time,
            phase.sim.time,
            phase.sim.iCa,
            phase.p.visual.single_cell_index,
            fig=None,
            ax=None,
            lncolor='g',
            ionname='Ca2+ cell',
        )
        axA.set_title('Cytosolic Ca2+ in cell {}'.format(
            phase.p.visual.single_cell_index))

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='cytosol_Ca_time')


    @piperunner(
        categories=('Ion Concentration', 'Calcium', 'Endoplasmic Reticulum',),
        requirements=phasereqs.ION_CALCIUM_DYNAMICS,
    )
    def export_ion_calcium_er(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot all calcium (i.e., Ca2+) ion concentrations for the endoplasmic
        reticulum of the single cell indexed by the current simulation
        configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        # One-dimensional Numpy array of all ion concentrations for the
        # endoplasmic reticulum of this cell over all sampled time steps.
        times_cell_ion_calcium_er = [
            cells_ion_calcium_er[phase.p.visual.single_cell_index]
            for cells_ion_calcium_er in phase.sim.endo_retic.Ca_er_time
        ]

        # Plot this array.
        pyplot.figure()
        pyplot.plot(phase.sim.time, times_cell_ion_calcium_er)

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='CaER')


    @piperunner(
        categories=('Voltage', 'Transmembrane', 'Endoplasmic Reticulum',),
        requirements=phasereqs.ION_CALCIUM_DYNAMICS,
    )
    def export_voltage_membrane_er(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot all transmembrane voltages across the endoplasmic reticulum of the
        single cell indexed by the current simulation configuration over all
        sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        # One-dimensional Numpy array of all ion concentrations for the
        # endoplasmic reticulum of this cell over all sampled time steps.
        times_cell_vmem_er = [
            cells_vmem_er[phase.p.visual.single_cell_index]
            for cells_vmem_er in phase.sim.endo_retic.ver_time
        ]

        # Plot this array.
        pyplot.figure()
        pyplot.plot(phase.sim.time, times_cell_vmem_er)

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='VER')

    # ..................{ EXPORTERS ~ pressure              }..................
    @piperunner(
        categories=('Pressure', 'Osmotic',),
        requirements=phasereqs.PRESSURE_OSMOTIC,
    )
    def export_pressure_osmotic(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot the osmotic cellular pressure for the single cell indexed by the
        current simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        p_osmo = tuple(
            arr[phase.p.visual.single_cell_index]
            for arr in phase.sim.osmo_P_delta_time)

        pyplot.figure()
        axOP = pyplot.subplot(111)
        axOP.plot(phase.sim.time, p_osmo)
        axOP.set_xlabel('Time [s]')
        axOP.set_ylabel('Osmotic Pressure [Pa]')
        axOP.set_title(
            'Osmotic pressure in cell {}'.format(phase.p.visual.single_cell_index))

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='OsmoticP_time')


    @piperunner(
        categories=('Pressure', 'Total',),
        requirements=phasereqs.PRESSURE_TOTAL,
    )
    def export_pressure_total(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot the **total cellular pressure** (i.e., summation of the cellular
        mechanical and osmotic pressure) for the single cell indexed by the
        current simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        p_hydro = tuple(
            arr[phase.p.visual.single_cell_index]
            for arr in phase.sim.P_cells_time)

        pyplot.figure()
        axOP = pyplot.subplot(111)
        axOP.plot(phase.sim.time, p_hydro)
        axOP.set_xlabel('Time [s]')
        axOP.set_ylabel('Total Pressure [Pa]')
        axOP.set_title(
            'Total pressure in cell {}'.format(phase.p.visual.single_cell_index))

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='P_time')

    # ..................{ EXPORTERS ~ pump                  }..................
    #FIXME: Actually plot the average of these rates. Currently, this method
    #only plots rates for a single arbitrarily selected membrane of this cell.
    @piperunner(
        categories=('Pump Rate', 'Na-K-ATPase',),
        requirements=phasereqs.PUMP_NAKATPASE,
    )
    def export_pump_nakatpase(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot the averages of the Na-K-ATPase membrane pump rates for the single
        cell indexed by the current simulation configuration over all time
        steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        pyplot.figure()
        axNaK = pyplot.subplot()

        if phase.p.is_ecm:
            #FIXME: Generalize to average across all membranes of this cell.
            pump_array_index = phase.cells.cell_to_mems[
                phase.p.visual.single_cell_index][0]
        else:
            pump_array_index = phase.p.visual.single_cell_index

        pump_rate = tuple(
            pump_array[pump_array_index]
            for pump_array in phase.sim.rate_NaKATP_time)

        axNaK.plot(phase.sim.time, pump_rate)
        axNaK.set_xlabel('Time [s]')
        axNaK.set_ylabel('Pumping Rate of Na+ Out of Cell [mol/(m2 s)]')
        axNaK.set_title(
            'Rate of NaK-ATPase pump in cell: {}'.format(
                phase.p.visual.single_cell_index))

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='NaKATPase_rate')

    # ..................{ EXPORTERS ~ voltage               }..................
    #FIXME: Actually plot the average of these Vmems. Currently, this method
    #only plots Vmems for a single arbitrarily selected membrane of this cell.
    @piperunner(categories=('Voltage', 'Transmembrane', 'Average',))
    def export_voltage_membrane(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot the averages of all transmembrane voltages for the single cell
        indexed by the current simulation configuration over all sampled time
        steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        #FIXME: Generalize to average across all membranes of this cell.
        mem_i = phase.cells.cell_to_mems[phase.p.visual.single_cell_index][0]

        # Plot single cell Vmem vs time.
        figVt, axVt = plotutil.plotSingleCellVData(
            phase.sim,
            mem_i,
            phase.p,
            fig=None, ax=None,
            lncolor='k',
        )
        axVt.set_title(
            'Voltage (Vmem) in cell {}'.format(phase.p.visual.single_cell_index))

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='Vmem_time')


    #FIXME: Actually plot the average of these Vmems. Currently, this method
    #only plots Vmems for a single arbitrarily selected membrane of this cell.
    @piperunner(categories=(
        'Voltage', 'Transmembrane', 'Fast Fourier Transform (FFT)',))
    def export_voltage_membrane_fft(
        self, phase: SimPhase, conf: SimConfExportPlotCell) -> None:
        '''
        Plot the fast Fourier transform (FFT) of the averages of all
        transmembrane voltages for the single cell indexed by the current
        simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        #FIXME: Generalize to average across all membranes of this cell.
        mem_i = phase.cells.cell_to_mems[phase.p.visual.single_cell_index][0]

        figFFT, axFFT = plotutil.plotFFT(
            phase.sim.time, phase.sim.vm_time, mem_i, lab="Power")
        axFFT.set_title('Fourier transform of Vmem in cell {}'.format(
            phase.p.visual.single_cell_index))

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='Vmem_FFT_time')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Facilities guaranteeing backward compatibility with prior file formats for
**pickled simulations** (i.e., previously pickled seed, initialization, and
simulation files).
'''

# ....................{ IMPORTS                           }....................
import sys
from betse.util.io.log import logs
# from betse.util.type.types import type_check

# ....................{ UPGRADERS                         }....................
def upgrade_sim_imports() -> None:
    '''
    Upgrade the in-memory module and class structure of the active Python
    interpreter to reflect the newest structure of these modules and classes
    expected by the current version of this application.

    This function preserves backward importability (and hence compatibility)
    with all prior supported pickled simulation formats, converting the
    obsolete module and class names imported by these formats into their modern
    equivalents. Specifically, for each obsolete module or class name imported
    by a prior supported pickled simulation format, this function injects an
    in-memory alias mapping from the obselete to modern such name.

    Since the :mod:`pickle` API provides no explicit means of doing so, this
    function modifies the in-memory module and class structure of the active
    Python interpreter *before* the :mod:`pickle` API is invoked to deserialize
    pickled simulation files. Failing to call this function *before*
    deserializing simulation files pickled by older versions of this
    application reliably induces the :mod:`pickle` API to raise obscure and
    non-human-readable :class:`ImportError` exceptions.

    Ideally, leveraging the third-party :mod:`dill` dependency would
    automatically resolve such backward importability issues by transitively
    pickling all imports required by pickled files in those files. Sadly, this
    does *not* currently appear to be the case.

    See Also
    ----------
    :func:`betse.science.config.confcompat.upgrade_sim_conf`
        Further details on which prior formats exactly are supported.
    '''

    # Upgrade package imports to each successive format. For safety, each
    # upgrade is performed in strict chronological order.
    _upgrade_sim_imports_to_0_5_2()
    _upgrade_sim_imports_to_0_6_0()
    _upgrade_sim_imports_to_0_7_1()
    _upgrade_sim_imports_to_0_9_3()

# ....................{ UPGRADERS ~ 0.5.2                 }....................
def _upgrade_sim_imports_to_0_5_2() -> None:
    '''
    Upgrade the in-memory module and class structure of the active Python
    interpreter to reflect the newest structure of these modules and classes
    expected by version 0.5.2 (i.e., "Happiest Hodgkin") of this application.
    '''

    # Log this upgrade attempt.
    logs.log_debug('Upgrading simulation imports to 0.5.2 format...')

    # Import all modules whose fully-qualified names have been modified.
    from betse.lib.yaml.abc import yamlabc, yamllistabc
    from betse.science import channels
    from betse.science.math import finitediff
    from betse.science.phase import phasecls
    from betse.science.config.export.visual import (
        confexpvisanim, confexpvisplot, confexpvisabc)
    from betse.science.config.export.visual.confexpvisanim import (
        SimConfExportAnimCells, SimConfExportAnimCellsEmbedded)
    from betse.science.config.export.visual.confexpvisplot import (
        SimConfExportPlotCells)
    from betse.util.type.iterable.mapping import mapcls

    # Alias obsolete module names to current module objects.
    sys.modules['betse.science.config.confabc'] = yamlabc
    sys.modules['betse.science.config.export.visual.confvisualabc'] = (
        confexpvisabc)
    sys.modules['betse.science.finitediff'] = finitediff
    sys.modules['betse.science.tissue.channels'] = channels
    sys.modules['betse.science.plot.plotconfig'] = confexpvisplot
    sys.modules['betse.science.plot.anim.animconfig'] = confexpvisanim
    sys.modules['betse.science.visual.anim.animconfig'] = confexpvisanim
    sys.modules['betse.science.visual.plot.plotconfig'] = confexpvisplot
    sys.modules['betse.util.type.mappings'] = mapcls

    # Alias obsolete to current class names.
    yamlabc.SimConfList = yamllistabc.YamlList
    confexpvisanim.SimConfAnimOne       = SimConfExportAnimCells
    confexpvisabc.SimConfVisualABC      = confexpvisabc.SimConfVisualCellsABC
    confexpvisabc.SimConfVisualMixin    = confexpvisabc.SimConfVisualCellsYAMLMixin
    confexpvisabc.SimConfVisualMolecule = confexpvisabc.SimConfVisualCellsNonYAML
    confexpvisabc.SimConfVisualGeneric  = SimConfExportAnimCellsEmbedded
    confexpvisabc.SimConfVisualListable = SimConfExportPlotCells
    confexpvisabc.SimConfVisual         = SimConfExportPlotCells
    confexpvisabc.SimConfListableVisual = SimConfExportPlotCells
    phasecls.SimPhaseType = phasecls.SimPhaseKind
    sys.modules['betse.science.config.export.visual.confexpvisanim'].SimConfAnim = (
        confexpvisanim.SimConfExportAnims)
    sys.modules['betse.science.config.export.visual.confexpvisplot'].SimConfPlot = (
        confexpvisplot.SimConfExportPlots)
    sys.modules['betse.science.visual.anim.animconfig'].AnimConfig = (
        confexpvisanim.SimConfExportAnims)
    sys.modules['betse.science.visual.plot.plotconfig'].PlotConfig = (
        confexpvisplot.SimConfExportPlots)

# ....................{ UPGRADERS ~ 0.6.x                 }....................
def _upgrade_sim_imports_to_0_6_0() -> None:
    '''
    Upgrade the in-memory module and class structure of the active Python
    interpreter to reflect the newest structure of these modules and classes
    expected by version 0.6.0 of this application.
    '''

    # Log this upgrade attempt.
    logs.log_debug('Upgrading simulation imports to 0.6.0 format...')

    # Import all modules whose fully-qualified names have been modified.
    from betse.lib.yaml.abc import yamlabc
    from betse.science.config.model import conftis
    from betse.science.config.export.visual import (
        confexpvisanim, confexpvisplot, confexpvisabc)
    from betse.science.tissue import tisprofile, tishandler
    from betse.science.tissue.event import tisevecut, tisevevolt
    from betse.science.tissue.picker import tispickcls, tispickimage
    from betse.science.tissue.picker.tispickcls import TissuePickerPercent
    from betse.science.tissue.picker.tispickimage import (
        TissuePickerImage, TissuePickerImageMask)
    from betse.science.tissue.tisprofile import CutProfile

    # Alias obsolete module names to current module objects.
    sys.modules['betse.lib.yaml.yamlabc'] = yamlabc
    sys.modules['betse.science.config.event.eventcut'] = tisevecut
    sys.modules['betse.science.config.event.eventvoltage'] = tisevevolt
    sys.modules['betse.science.config.export.confexpvisanim'] = confexpvisanim
    sys.modules['betse.science.config.export.confexpvisplot'] = confexpvisplot
    sys.modules['betse.science.config.export.confvis'] = confexpvisabc
    sys.modules['betse.science.config.tissue.conftis'] = conftis
    sys.modules['betse.science.tissue.bitmapper'] = tispickimage
    sys.modules['betse.science.tissue.handler'] = tishandler
    sys.modules['betse.science.tissue.tiscls'] = tisprofile
    sys.modules['betse.science.tissue.tissuecls'] = tisprofile
    sys.modules['betse.science.tissue.tissuepick'] = tispickcls

    # Alias obsolete to current class names.
    tispickcls.TissuePickerBitmap = TissuePickerImage
    tispickcls.TissuePickerRandom = TissuePickerPercent
    tispickimage.BitMapper = TissuePickerImageMask
    tisprofile.TissueCut = CutProfile

# ....................{ UPGRADERS ~ 0.7.x                 }....................
def _upgrade_sim_imports_to_0_7_1() -> None:
    '''
    Upgrade the in-memory module and class structure of the active Python
    interpreter to reflect the newest structure of these modules and classes
    expected by version 0.7.1 of this application.
    '''

    # Log this upgrade attempt.
    logs.log_debug('Upgrading simulation imports to 0.7.1 format...')

    # Import all modules whose fully-qualified names have been modified.
    from betse.science import channels
    from betse.science.config.export import visual

    # Alias obsolete module names to current module objects.
    sys.modules['betse.science.channelo'] = channels
    sys.modules['betse.science.config.visual'] = visual

# ....................{ UPGRADERS ~ 0.9.x                 }....................
def _upgrade_sim_imports_to_0_9_3() -> None:
    '''
    Upgrade the in-memory module and class structure of the active Python
    interpreter to reflect the newest structure of these modules and classes
    expected by version 0.9.3 of this application.
    '''

    # Log this upgrade attempt.
    logs.log_debug('Upgrading simulation imports to 0.9.3 format...')

    # Import all modules whose fully-qualified names have been modified.
    from betse.science.config.export import confexpcsv
    from betse.science.config.export.visual import (
        confexpvisabc, confexpvisanim, confexpvisplot)
    from betse.science.config.export.visual.confexpvisanim import (
        SimConfExportAnims, SimConfExportAnimCellsEmbedded)
    from betse.science.config.export.visual.confexpvisplot import (
        SimConfExportPlots, SimConfExportPlotCell, SimConfExportPlotCells)
    from betse.science.enum import enumphase
    from betse.util.type.iterable import mapping

    # Alias obsolete module names to current module objects.
    sys.modules['betse.util.type.mapping'] = mapping
    sys.modules['betse.science.config.export.confcsv'] = confexpcsv
    sys.modules['betse.science.config.export.visual.confanim'] = confexpvisanim
    sys.modules['betse.science.config.export.visual.confplot'] = confexpvisplot
    sys.modules['betse.science.config.export.visual.confvisabc'] = (
        confexpvisabc)
    sys.modules['betse.science.phase.phaseenum'] = enumphase

    # Alias obsolete to current class names.
    confexpvisabc.SimConfVisualCellListItem  = SimConfExportPlotCell
    confexpvisabc.SimConfVisualCellsListItem = SimConfExportPlotCells
    confexpvisabc.SimConfVisualCellsEmbedded = SimConfExportAnimCellsEmbedded
    confexpvisanim.SimConfAnimAll = SimConfExportAnims
    confexpvisplot.SimConfPlotAll = SimConfExportPlots



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Concrete subclasses defining this application's command line interface (CLI).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on application startup, the
# top-level of this module may import *ONLY* from submodules guaranteed to:
# * Exist, including standard Python and application modules.
# * Never raise exceptions on importation (e.g., due to module-level logic).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from betse.cli import cliinfo
from betse.util.cli.clicmd import (
    CLISubcommander,
    CLISubcommandNoArg,
    CLISubcommandParent,
    CLISubcommandYAMLOnly,
)
from betse.util.cli.clicmdabc import CLISubcommandableABC
from betse.util.cli.cliopt import CLIOptionBoolTrue
from betse.util.io.log import logs
from betse.util.os import displays
from betse.util.path import files, pathnames
from betse.util.type.decorator.decmemo import property_cached
from betse.util.type.types import SequenceTypes

# ....................{ SUBCLASS                           }....................
class BetseCLI(CLISubcommandableABC):
    '''
    Command line interface (CLI) for this application.

    Attributes (of :attr:`_args`)
    ----------
    is_headless : bool
        ``True`` only if the active Python interpreter is to be coerced into
        running **headless** (i.e., with *no* access to a GUI display).
        Defaults to ``False``.
    '''

    # ..................{ SUPERCLASS ~ property : commands   }..................
    @property_cached
    def _subcommander_top(self) -> CLISubcommander:

        return CLISubcommander(
            subcommand_var_name='subcommand_name_top',
            help_title='subcommands',
            subcommands=(

                CLISubcommandYAMLOnly(
                    name='config',
                    help_synopsis=(
                        'create a default config file for '
                        '{program_name} simulations'
                    ),
                    help_description='''
Write a default tissue simulation configuration to the passed output file. While
not strictly necessary, this file should have filetype ".yaml" . If this file
already exists, an error will be printed.

You may edit this file at any time. By default, this file instructs
{program_name} to save simulation results (e.g., plots) to the directory
containing this file.
''',),


                CLISubcommandYAMLOnly(
                    name='seed',
                    help_synopsis='seed a new cell cluster for a config file',
                    help_description='''
Create the cell cluster defined by the passed configuration file. The results
will be saved to output files defined by this configuration.
''',),


                CLISubcommandYAMLOnly(
                    name='init',
                    help_synopsis=(
                        'initialize a seeded cell cluster for a config file'),
                    help_description='''
Initialize (i.e., calculate steady-state concentrations for) the previously
created cell cluster defined by the passed configuration file. Initialization
results will be saved to output files defined by this configuration, while the
previously created cell cluster will be loaded from input files defined by this
configuration.
''',),


                CLISubcommandYAMLOnly(
                    name='sim',
                    help_synopsis=(
                        'simulate an initialized cell cluster '
                        'for a config file'
                    ),
                    help_description='''
Simulate the previously initialized cell cluster defined by the passed
configuration file. Simulation results will be saved to output files defined by
this configuration, while the previously initialized cell cluster will be loaded
from input files defined by this configuration.
''',),


                CLISubcommandYAMLOnly(
                    name='sim-grn',
                    help_synopsis=(
                        'simulate a gene regulatory network for a config file'),
                    help_description='''
Simulate a gene regulatory network (GRN) for the previously initialized cell
cluster defined by the passed configuration file, whose "gene regulatory network
config" option specifies the path of the configuration file defining this
network. All other simulation features and options will be ignored.

Simulation results will be saved to output files defined by this configuration,
while the previously initialized cell cluster will be loaded from input files
defined by this configuration.
''',),


                CLISubcommandParent(
                    name='plot',
                    help_synopsis=(
                        'plot an initialized or simulated simulation'),
                        # 'plot a seeded, initialized, or simulated simulation'),
                    help_description='''
Run the passed plotting subcommand. For example, to plot the previous
simulation defined by a configuration file "my_sim.yaml" in the current
directory:

;    betse plot sim my_sim.yaml
''',
                    subcommander=self._subcommander_plot,
                ),


                CLISubcommandNoArg(
                    name='info',
                    help_synopsis=(
                        'print metadata synopsizing '
                        '{program_name} and current system'
                    ),
                    help_description='''
Print informational metadata in ":"-delimited key-value format, including:

* Program name, version, and principal authors.

* Absolute paths of critical files and directories used by {program_name},
including:

* {program_name}'s data directory (i.e., the program-specific directory to
    which non-Python files intended for use by external users are stored).

* {program_name}'s dot directory (i.e., the user-specific directory to which
    files and directories intended for internal program use are stored).

* {program_name}'s log file (i.e., the user-specific file to which all runtime
    messages are appended, including low-level debug statements, non-fatal
    warnings, and fatal errors).
''',),


                CLISubcommandNoArg(
                    name='try',
                    help_synopsis=(
                        'create, init, simulate, and plot a sample simulation'),
                    help_description='''
Run a sample tissue simulation. This subcommand (A) creates a default YAML
configuration file, (B) creates the cell cluster defined by that file, (C)
initializes that cell cluster, (D) plots that initialization, (E) simulates that
initialization, and (F) plots that simulation. All files and directories created
by these operations will be preserved (rather than deleted on subcommand
completion).

Equivalently, this subcommand is shorthand for the following:

;    betse config    sample_sim/sample_sim.yaml
;    betse seed      sample_sim/sample_sim.yaml
;    betse init      sample_sim/sample_sim.yaml
;    betse sim       sample_sim/sample_sim.yaml
;    betse plot init sample_sim/sample_sim.yaml
;    betse plot sim  sample_sim/sample_sim.yaml
''',),

            ))


    @property
    def _subcommander_plot(self) -> CLISubcommander:
        '''
        Container of all child subcommands accepted by the ``plot`` subcommand.
        '''

        # Return a container of all child subcommands accepted by this parent
        # "plot" subcommand.
        return CLISubcommander(
            subcommand_var_name='subcommand_name_plot',
            help_title='plot subcommands',
            subcommands=(
                CLISubcommandYAMLOnly(
                    name='seed',
                    help_synopsis=(
                        'plot a seeded cell cluster for a config file'),
                    help_description='''
Plot the previously seeded cell cluster defined by the passed configuration
file. Plot results will be saved to output files defined by this configuration,
while the previously seeded cell cluster will be loaded from input files
defined by this configuration.
''',),


                CLISubcommandYAMLOnly(
                    name='init',
                    help_synopsis=(
                        'plot an initialized cell cluster for a config file'),
                    help_description='''
Plot the previously initialized cell cluster defined by the passed configuration
file. Plot results will be saved to output files defined by this configuration,
while the previously initialized cell cluster will be loaded from input files
defined by this configuration.
''',),


                CLISubcommandYAMLOnly(
                    name='sim',
                    help_synopsis=(
                        'plot a simulated cell cluster for a config file'),
                    help_description='''
Plot the previously simulated cell cluster defined by the passed configuration
file. Plot results will be saved to output files defined by this configuration,
while the previously simulated cell cluster will be loaded from input files
defined by this configuration.
''',
            ),


                CLISubcommandYAMLOnly(
                    name='sim-grn',
                    help_synopsis=(
                        'plot a simulated gene regulatory network '
                        'for a config file'
                    ),
                    help_description='''
Plot the previously simulated gene regulatory network (GRN) defined by the
passed configuration file. Plot results will be saved to output files defined by
this configuration, while the previously simulated cell cluster will be loaded
from input files defined by this configuration.
''',),

            ))

    # ..................{ SUPERCLASS ~ property : options    }..................
    @property
    def _options_top(self) -> SequenceTypes:

        # Sequence of all default top-level options.
        options_top = super()._options_top

        # Return a list extending this sequence with subclass-specific options.
        return options_top + [
            CLIOptionBoolTrue(
                long_name='--headless',
                synopsis=(
                    'enable headless mode '
                    '(display no plots or animations) '
                    '[forces "--matplotlib-backend=agg"]'
                ),
            ),
        ]


    def _parse_options_top(self) -> None:

        # Parse all default top-level options.
        super()._parse_options_top()

        # If coercing this Python interpreter into running headless, do so.
        if self._args.is_headless:
            displays.set_headless(True)
        # Else, allow the displays_is_headless() function to implicitly detect
        # whether this interpreter is actually running headless.

    # ..................{ SUPERCLASS ~ help                  }..................
    def _log_header(self) -> None:

        cliinfo.log_header()

    # ..................{ PROPERTIES                         }..................
    @property_cached
    def _sim_runner(self) -> 'betse.science.simrunner.SimRunner':
        '''
        Simulation runner running simulation subcommands on the YAML-formatted
        simulation configuration file passed by the user at the CLI.
        '''

        # Defer heavyweight imports.
        from betse.science.parameters import Parameters
        from betse.science.simrunner import SimRunner

        # Simulation configuration loaded from this YAML-formatted file.
        p = Parameters.make(conf_filename=self._args.conf_filename)

        # Create and return a simulation runner for this configuration.
        return SimRunner(p=p)

    # ..................{ SUBCOMMANDS                        }..................
    def _do_info(self) -> None:
        '''
        Run the ``info`` subcommand.
        '''

        cliinfo.log_info()

    # ..................{ SUBCOMMANDS ~ sim                  }..................
    def _do_try(self) -> object:
        '''
        Run the ``try`` subcommand and return the result of doing so.
        '''

        # Basename of the sample configuration file to be created.
        config_basename = 'sample_sim.yaml'

        # Relative filename of this file, relative to the current directory.
        self._args.conf_filename = pathnames.join(
            'sample_sim', config_basename)

        #FIXME: Insufficient. We only want to reuse this file if this file's
        #version is identical to that of the default YAML configuration file's
        #version. Hence, this logic should (arguably) be shifted elsewhere --
        #probably into "betse.science.sim_config".

        # If this file already exists, reuse this file.
        if files.is_file(self._args.conf_filename):
            logs.log_info(
                'Reusing simulation configuration "%s".', config_basename)
        # Else, create this file.
        else:
            self._do_config()

        # Run all general-purposes phases, thus excluding network-isolated
        # phases (e.g., "_do_sim_grn"), in the expected order.
        self._do_seed()
        self._do_init()
        self._do_sim()
        self._do_plot_seed()
        self._do_plot_init()

        # Return the value returned by the last such phase, permitting this
        # subcommand to be memory profiled. While any value would technically
        # suffice, the value returned by the last such phase corresponds to a
        # complete simulation run and hence is likely to consume maximal memory.
        return self._do_plot_sim()


    def _do_config(self) -> None:
        '''
        Run the ``config`` subcommand.
        '''

        # Defer heavyweight imports.
        from betse.science.parameters import Parameters

        # Copy the default simulation configuration to this target file.
        Parameters().copy_default(trg_conf_filename=self._args.conf_filename)


    def _do_seed(self) -> object:
        '''
        Run the ``seed`` subcommand and return the result of doing so.
        '''

        return self._sim_runner.seed()


    def _do_init(self) -> object:
        '''
        Run the ``init`` subcommand and return the result of doing so.
        '''

        return self._sim_runner.init()


    def _do_sim(self) -> object:
        '''
        Run the ``sim`` subcommand and return the result of doing so.
        '''

        return self._sim_runner.sim()


    def _do_sim_grn(self) -> object:
        '''
        Run the ``sim-grn`` subcommand and return the result of doing so.
        '''

        return self._sim_runner.sim_grn()


    def _do_plot_seed(self) -> object:
        '''
        Run the ``plot`` subcommand's ``seed`` subcommand and return the result
        of doing so.
        '''

        return self._sim_runner.plot_seed()


    def _do_plot_init(self) -> object:
        '''
        Run the ``plot`` subcommand's ``init`` subcommand and return the result
        of doing so.
        '''

        return self._sim_runner.plot_init()


    def _do_plot_sim(self) -> object:
        '''
        Run the ``plot`` subcommand's ``sim`` subcommand and return the result
        of doing so.
        '''

        return self._sim_runner.plot_sim()


    def _do_plot_sim_grn(self) -> object:
        '''
        Run the ``plot`` subcommand's ``sim-grn`` subcommand and return the
        result of doing so.
        '''

        return self._sim_runner.plot_grn()



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Layer subclasses spatially overlaying streamlines as stream plots onto the
current cell cluster.
'''

#FIXME: Optimize. The current _get_time_currents() approach is extremely
#inefficient, as this entire vector field will be recomputed for each plot and
#animation when current overlays are enabled. The simpliest alternative is to
#add two new properties to the "Simulator" class:
#
#* Simulator.time_currents_intra(), returning the vector field...
#FIXME: Oh, wait. These vector fields require the "sim", "cells", and "p"
#parameters -- which, due to arguably poor Simulator design, are currently
#unavailable as attributes there. Very well. We see little choice but to
#finally define a pipeline class. As these vector fields are currently only
#required by plots and animations, sequestering these fields to that pipeline
#class would be trivial. However, we would then need to pass the same instance
#of that class to *EVERY* instantiated plot and animation.
#
#Doing so is ultimately trivial but tedious and hence deferred to another day.

# ....................{ IMPORTS                           }....................
import numpy as np
from betse.science.math import mathunit
from betse.science.visual.layer.vectorfield.lyrvecfldabc import (
    LayerCellsFieldColorlessABC)
# from betse.util.type.types import type_check
from matplotlib.patches import FancyArrowPatch

# ....................{ SUBCLASSES                        }....................
class LayerCellsFieldStream(LayerCellsFieldColorlessABC):
    '''
    Layer subclass both simulating *and* plotting streamlines of a single
    vector field (e.g., total current density) onto the cell cluster for one on
    more simulation time steps.

    This layer is somewhat more computationally expensive in both space and
    time than the average layer. For each plot or animation frame to be layered
    with streamlines, an internal fluid simulation of the density of the
    desired vector field through the cell cluster specific to this frame is
    solved.

    Attributes
    ----------
    _stream_plot : matplotlib.streamplot.StreamplotSet
        Streamplot of all streamlines previously plotted for the prior time
        step if any *or* ``None`` otherwise, temporarily preserved for only one
        time step to permit its removal prior to plotting a new streamplot for
        the current time step.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass.
        super().__init__(*args, **kwargs)

        # Default all remaining instance variables.
        self._stream_plot = None

    # ..................{ SUPERCLASS                        }..................
    def _layer_first(self) -> None:
        '''
        Simulate and layer streamlines of this vector field for the first time
        step onto the figure axes of the current plot or animation.
        '''

        # Arrays of the upscaled X and Y coordinates of all grid spaces.
        grid_x = mathunit.upscale_coordinates(self._phase.cells.X)
        grid_y = mathunit.upscale_coordinates(self._phase.cells.Y)

        # Vector field whose X and Y components are spatially situated at grid
        # space centres.
        field = self._field.times_grids_centre

        # Arrays of all magnitudes *AND* normalized X and Y components of this
        # vector field for this time step.
        field_magnitudes = field.magnitudes[self._visual.time_step]
        field_unit_x = field.unit_x[self._visual.time_step]
        field_unit_y = field.unit_y[self._visual.time_step]

        # Maximum magnitude of this vector field for this time step.
        field_magnitude_max = np.max(field_magnitudes)

        # One-dimensional array of the visual widths of all vectors of this
        # vector field for this time step.
        streamlines_width = (
            3.0 * field_magnitudes / field_magnitude_max) + 0.5

        # Streamplot of all streamlines plotted for this time step. See the
        # matplotlib.streamplot.streamplot() docstring for further details.
        self._stream_plot = self._visual.axes.streamplot(
            # X and Y coordinates of all grid points.
            x=grid_x,
            y=grid_y,

            # X and Y normalized components of this vector field.
            u=field_unit_x,
            v=field_unit_y,

            # Matplotlib-specific color code of all streamlines.
            color=self._phase.p.vcolor,

            # Density of streamlines in both the X and Y dimensions.
            density=self._phase.p.stream_density,

            # Line widths of all streamlines.
            linewidth=streamlines_width,

            #FIXME: For still frames, an arrow size of about 5.0 is best; for
            #rendered video, these blow up and a size of 3.0 is better. Not that
            #this is important enough to ever get to it. :)

            # Factor by which to upscale the size of all streamline arrowheads.
            # arrowsize=3.0,

            # Z-order of these streamlines with respect to other artists.
            zorder=self._zorder,
        )


    def _layer_next(self) -> None:
        '''
        Simulate and layer streamlines of this vector field for the next time
        step onto the figure axes of the current plot or animation.
        '''

        # Remove all streamlines plotted for the prior time step.
        self._stream_plot.lines.remove()

        # If this matplotlib version supports removing the set of all streamline
        # arrowheads plotted for the prior time step, do so.
        try:
            self._stream_plot.arrows.remove()
        # Else, these arrowheads *MUST* be manually erased by iterating over all
        # patch objects and preserving all non-arrowhead patches. Doing so also
        # removes all arrowhead patches of other visuals already plotted for
        # this time step and is hence non-ideal. But no alternatives exist.
        except NotImplementedError:
            #FIXME: This simplistic approach *USED* to work. For unknown and
            #presumably indefensible reason, some undisclosed newer version of
            #matplotlib silently broke backward compatibility by preventing this
            #from working. This now raises the unreadable exception:
            #    AttributeError: can't set attribute 'patches'
            #Sadly, no one else has reported a similar issue on the matplotlib
            #issue tracker. We thus have no recourse but to attempt a completely
            #different workaround now. *sigh*
            # self._visual.axes.patches = [
            #     patch
            #     for patch in self._visual.axes.patches
            #     if not isinstance(patch, FancyArrowPatch)
            # ]

            # Rather than attempting to replace the entire axes patches, instead
            # selectively search for and remove *ALL* arrow-specific axes
            # patches. Again, this is highly non-ideal. It is what it is.
            #
            # This workaround is strongly inspired by this StackOverflow answer:
            #     https://stackoverflow.com/a/61932726/5049231
            for artist in self._visual.axes.get_children():
                if isinstance(artist, FancyArrowPatch):
                    artist.remove()

        # Replot this streamplot for this time step.
        self._layer_first()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Fixtures and fixture classes creating temporary simulation configurations
isolated to specific tests, which typically modify the contents of these
configurations so as to exercise specific feature sets and edge cases.
'''

# ....................{ IMPORTS                            }....................
from betse_test._fixture import initter
from betse_test._fixture.simconf.simconfclser import SimConfTestInternal
from pytest import fixture

# ....................{ FIXTURES ~ default                 }....................
@fixture
def betse_sim_conf_default(
    betse_temp_dir: 'py._path.local.LocalPath') -> SimConfTestInternal:
    '''
    Per-test fixture creating a temporary default simulation configuration file
    and returning a wrapper around this file.

    Unlike the minified simulation configuration created by the
    :func:`betse_sim_conf` fixture and leveraged by most tests, the default
    simulation configuration created by this fixture remains unmodified (except
    for unavoidably disabling interactive simulation features, which
    non-interactive testing requires). Tests leveraging this fixture incur a
    significant performance penalty but can expose edge-case issues obscured by
    minification, including computational instability produced by the default
    non-minified time steps.

    Parameters
    ----------
    betse_temp_dir : py._path.local.LocalPath
        Wrapper around a temporary directory isolated to the current test.

    Returns
    ----------
    SimConfTestInternal
        Wrapper around a temporary default simulation configuration file.

    See Also
    ----------
    :func:`betse_sim_conf`
        Further details, ignoring minification performed by this fixture.
    '''

    # Defer heavyweight imports.
    from betse.science.parameters import Parameters

    # Initialize the application metadata singleton, as required by the
    # subsequent access of the "Parameters.conf_default_filename" class
    # property.
    initter.init_app()

    # Wrapper wrapping the default simulation configuration file copied into
    # this temporary directory and sanitized therein.
    sim_state = SimConfTestInternal(
        src_conf_filename=Parameters.conf_default_filename,
        trg_conf_filepath=betse_temp_dir.join('sim_config.yaml'),
    )

    # Return this wrapper *WITHOUT* minifying this configuration.
    return sim_state

# ....................{ FIXTURES ~ minified                }....................
@fixture
def betse_sim_conf(
    betse_sim_conf_default: SimConfTestInternal) -> SimConfTestInternal:
    '''
    Per-test fixture creating a temporary minified simulation configuration file
    and returning a wrapper around this file.

    Configuration Modifications (On-disk)
    ----------
    This fixture copies the default simulation configuration file for this
    application -- complete with all external assets (e.g., geometry masks)
    referenced and required by this file -- into a temporary directory whose
    basename is the name of the test requesting this fixture excluding the
    prefixing substring ``test_``. When requested by the
    ``test_cli_sim_default`` test, for example, this fixture creates a temporary
    simulation configuration file ``{tmpdir}/cli_sim_default/sim_config.yaml``
    for the absolute path ``{tmpdir}`` of this test session's root temporary
    directory (e.g., ``/tmp/pytest-0/cli_sim_default/sim_config.yaml``).

    This directory and thus simulation configuration is safely accessible *only*
    for the duration of the current test. Subsequently run tests and fixtures
    *cannot* safely reuse this configuration.

    Configuration Modifications (In-memory)
    ----------
    This fixture also transforms the in-memory instance of the
    :class:`betse.science.parameters.Parameters` class encapsulating this
    configuration as follows:

    * All configuration options either requiring interactive input *or*
      displaying interactive output are disabled (e.g., plots, animations).
    * The space and time costs associated with simulating this configuration
      are safely minimized in a manner preserving all features.

    Since this fixture writes these changes back to this file, the parent
    fixture or test is *not* required to do so manually (e.g., by calling the
    :meth:`SimConfTestInternal.p.save_inplace` method on the object returned by
    this fixture).

    Parameters
    ----------
    betse_sim_conf_default : SimConfTestInternal
        Wrapper around a temporary non-minified simulation configuration file.

    Returns
    ----------
    SimConfTestInternal
        Wrapper around a temporary simulation configuration file specific to the
        current test, including such metadata as:

        * The absolute path of this configuration's on-disk YAML file.
        * This configuration's in-memory dictionary deserialized from this file.
    '''

    # Minimize the space and time costs associated with this configuration.
    betse_sim_conf_default.config.minify()

    # Save these changes back to the same file.
    betse_sim_conf_default.p.save_inplace()

    # Return this wrapper.
    return betse_sim_conf_default

# ....................{ FIXTURES ~ compatible              }....................
@fixture
def betse_sim_conf_compat(
    betse_temp_dir: 'py._path.local.LocalPath') -> SimConfTestInternal:
    '''
    Per-test fixture creating and returning a wrapper around a temporary
    simulation configuration file (complete with a pickled seed, initialization,
    and simulation) produced by the oldest version of this application for which
    the current version of this application guarantees backward compatibility.

    Caveats
    ----------
    Unlike the object returned by the comparable :func:`betse_sim_conf`
    fixture, the object returned by this fixture is *not* safely modifiable by
    the current version of this application. Doing so would invalidate the
    pickled files produced by the older version of this application, which
    would largely defeat the purpose of invoking this fixture.

    Parameters
    ----------
    betse_temp_dir : py._path.local.LocalPath
        Wrapper around a temporary directory isolated to the current test.

    Returns
    ----------
    SimConfTestInternal
        Wrapper around a temporary simulation configuration file specific to
        the current test, complete with pickled seed, initialization, and
        simulation files produced by the older version of this application.
    '''

    # Defer heavyweight imports.
    from betse import metadata
    from betse.util.app.meta import appmetaone
    from betse.util.path import files

    # Initialize the application metadata singleton, as required by the
    # subsequent instantiation of the "SimConfTestInternal" subclass.
    initter.init_app()

    # Absolute dirname of the directory providing test data.
    test_data_dirname = appmetaone.get_app_meta().test_data_dirname

    # Absolute filename of the default simulation configuration file produced
    # by the oldest version of this application for which the current version
    # of this application guarantees backward compatibility.
    src_conf_filename = files.join_or_die(
        test_data_dirname,
        metadata.GIT_TAG_COMPAT_OLDEST,
        'yaml',
        'sim_config.yaml',
    )

    # Wrapper wrapping the default simulation configuration file copied into
    # this temporary directory and sanitized therein.
    sim_state = SimConfTestInternal(
        src_conf_filename=src_conf_filename,
        trg_conf_filepath=betse_temp_dir.join('sim_config.yaml'))

    # Return this wrapper *WITHOUT* minifying this configuration.
    return sim_state



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# ....................{ IMPORTS                            }....................
import math
import numpy as np
# import scipy.ndimage
from scipy.spatial import Delaunay, cKDTree

# ....................{ CLASSES                            }....................
class FiniteDiffSolver(object):
    '''
    Finite difference solver.

    This solver provides methods to establish a rectangular Marker and Cell
    (MACs) grid, compute discrete Laplacian operators and solve the Poisson,
    Laplacian, Heat, or Unsteady Stokes Flow (linearized Navier-Stokes)
    equations in 2D using the Central Finite Difference method.

    Attributes (Grid)
    ------------------------
    delta : NumericSimpleTypes
        Distance in meters between each grid point *and* between each grid
        space, uniformally applied in both the X and Y dimensions.

    Attributes (Grid Points)
    ----------
    Each grid point is a vertex equidistant from each adjacent grid points
    collectively forming the square lattice referred to as the grid.

    grid_nx : int
        Number of grid points in the X dimension.
    grid_ny : int
        Number of grid points in the Y dimension.
    verts_X : ndarray
        Two-dimensional Numpy array of the X coordinates of all grid points,
        whose:
        * First dimension indexes each row of this grid.
        * Second dimension indexes each column of this grid such that each
          element is the X coordinate in meters of the grid point at this row
          and column.
    verts_Y : ndarray
        Two-dimensional Numpy array of the Y coordinates of all grid points,
        whose:
        * First dimension indexes each row of this grid.
        * Second dimension indexes each column of this grid such that each
          element is the Y coordinate in meters of the grid point at this row
          and column.
    verts_shape : tuple
        2-tuple ``(height, width)`` of this grid, where:
        * ``height`` is the number of rows in this grid.
        * ``width`` is the number of columns in this grid.
    xmin : NumericSimpleTypes
        X coordinate in meters of the leftmost grid point(s).
    xmax : NumericSimpleTypes
        X coordinate in meters of the rightmost grid point(s).
    ymin : NumericSimpleTypes
        Y coordinate in meters of the bottom-most grid point(s).
    ymax : NumericSimpleTypes
        Y coordinate in meters of the topmost grid point(s).
    xy_verts : ndarray
        Two dimensional Numpy array of the Cartesian coordinates of all grid
        points, whose:
        * First dimension indexes each grid point in row-column order
          (beginning with the grid point at the origin in the lower-left corner
          and ending with the grid point at the upper-right corner, traversing
          grid points left to right).
        * Second dimension indexes each Cartesian coordinate of the current
          grid point, whose:
          * First element is the X coordinate of this grid point.
          * Second element is the Y coordinate of this grid point.

    Attributes (Grid Spaces)
    ----------
    Each grid space is the square whose four vertices are the four grid points
    surrounding that space such that a grid containing m x n grid points
    contains (m-1) x (n-1) grid spaces.

    cents_X : ndarray
        Two-dimensional Numpy array of the X coordinates of the centres of all
        grid spaces, whose:
        * First dimension indexes each row of this grid (excluding the last).
        * Second dimension indexes each column of this grid (excluding the
          last) such that each element is:
          * The X coordinate in meters of the centre of the grid space
            immediately above this row and to the right of this column.
          * Equivalently, the X coordinate of the midpoint between the grid
            points immediately to the left and right of this grid space.
    cents_X : ndarray
        Two-dimensional Numpy array of the Y coordinates of the centres of all
        grid spaces, whose:
        * First dimension indexes each row of this grid (excluding the last).
        * Second dimension indexes each column of this grid (excluding the
          last) such that each element is:
          * The Y coordinate in meters of the centre of the grid space
            immediately above this row and to the right of this column.
          * Equivalently, the Y coordinate of the midpoint between the grid
            points immediately below and above this grid space.
    cents_shape : tuple
        2-tuple ``(height, width)`` of the spaces of this grid, where:
        * ``height`` is the number of rows in this grid (excluding the last).
        * ``width`` is the number of columns in this grid (excluding the last).
        Hence, the following invariants are guaranteed to hold:
        * ``cents_shape[0] == verts_shape[0] - 1``.
        * ``cents_shape[1] == verts_shape[1] - 1``.
    xy_cents : ndarray
        Two dimensional Numpy array of the Cartesian coordinates of all grid
        spaces, whose:
        * First dimension indexes each grid space in row-column order
          (beginning with the grid space at the origin in the lower-left corner
          and ending with the grid space at the upper-right corner, traversing
          grid spaces left to right).
        * Second dimension indexes each Cartesian coordinate of the current
          grid space, whose:
          * First element is the X coordinate of this grid space.
          * Second element is the Y coordinate of this grid space.
    '''

    def __init__(self) -> None:
        pass

    def makeGrid(
        self, grid_n, xmin = -1, xmax = 1, ymin = -1, ymax = 1) -> None:
        """
        Create linear vectors, rectangular grids and expanded grids of spatial
        coordinates and solution vectors.This method produces a standard grid.

        delta:          spacing between grid points (same in the x and y dimensions)
        xmin:           minimum x-value  [m]
        xmax:           max x-value [m]
        ymin:           minimum y-value  [m]
        ymax:           max y value [m]
        """

        # Create linear vectors and rectangular grids of spatial coordinates (indexed by i and j)
        self.x = np.linspace(xmin,xmax,grid_n)
        self.y = np.linspace(ymin,ymax,grid_n)

        self.X,self.Y = np.meshgrid(self.x,self.y)

        self.delta_x = self.x[1] - self.x[0]
        self.delta_y = self.y[1] - self.y[0]

        # define a data structure that holds [x,y] coordinate points of each 2d grid-matrix entry
        self.xypts = []
        self.map_ij2k = []

        for i, row in enumerate(self.X):
            for j, val in enumerate(row):
                xpt = self.X[i,j]
                ypt = self.Y[i,j]
                self.xypts.append([xpt,ypt])
                self.map_ij2k.append([i,j])

        self.map_ij2k = np.asarray(self.map_ij2k)

    def cell_grid(self,grid_delta, xmin, xmax, ymin, ymax) -> None:
        """
        Make a staggered n x n grid for use with finite volume or MACs method.

        One n x n grid contains the cell centres.
        One n+1 x n+1 grid contains the cell corners (vertices)
        One n+1 x n grid contains the u-coordinate points for velocity and x-coordinate for fluxes
        One n x n+1 grid contains the v-coordinate points for velocity and y-coordinate for fluxes
        """

        # max-min dimensions of the 2D world space:
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax

        # spacing of the grid points (uniform in x and y dimensions):
        self.delta = grid_delta

        # number of points in the x and y directions
        self.grid_nx = int((self.xmax - self.xmin)/grid_delta)
        self.grid_ny = int((self.ymax - self.ymin)/grid_delta)

        # creation of the first mesh -- these are the corners of each MACs grid square:
        xv = np.linspace(xmin,xmax,self.grid_nx+1)
        yv = np.linspace(ymin,ymax,self.grid_ny+1)
        self.verts_X, self.verts_Y = np.meshgrid(xv,yv)

        x_verts = self.verts_X.ravel()
        y_verts = self.verts_Y.ravel()
        self.xy_verts = np.column_stack((x_verts,y_verts))
        self.verts_shape = self.verts_X.shape

        # grid and coordinate vectors defining the centre points of each cell:
        self.cents_X = (self.verts_X[0:-1,0:-1] + self.verts_X[0:-1,1:])/2
        self.cents_Y = (self.verts_Y[0:-1,0:-1] + self.verts_Y[1:,0:-1])/2

        x_cent = self.cents_X.ravel()
        y_cent = self.cents_Y.ravel()

        # unravelled x,y points for cell centres:
        self.xy_cents = np.column_stack((x_cent,y_cent))

        self.cents_shape = self.cents_X.shape

        # define a mapping between the (i,j) centers mesh and the linear unravelled k index:
        self.map_ij2k_cents = []

        for i, row in enumerate(self.cents_X):
            for j, val in enumerate(row):
                self.map_ij2k_cents.append([i,j])

        self.map_ij2k_cents = np.asarray(self.map_ij2k_cents)

        # Next define grids for the x and y cordinates as midpoints of each MACs cell:
        # cell side midpoints in the x (u) and x (v) directions:
        mx = self.cents_X[0,:]
        my = self.verts_Y[:,0]
        self.v_X, self.v_Y = np.meshgrid(mx,my)

        x_m = self.v_X.ravel()
        y_m = self.v_Y.ravel()

        # points on which y co-ordinate of velocity is defined:
        self.v_pts = np.column_stack((x_m,y_m))
        self.v_shape = self.v_X.shape

        self.map_ij2k_v = []

        for i, row in enumerate(self.v_X):
            for j, val in enumerate(row):
                self.map_ij2k_v.append([i,j])

        self.map_ij2k_v = np.asarray(self.map_ij2k_v)

        nx = self.verts_X[0,:]
        ny = self.cents_Y[:,0]
        self.u_X, self.u_Y = np.meshgrid(nx,ny)

        x_n = self.u_X.ravel()
        y_n = self.u_Y.ravel()

        # points on which x co-ordinate of velocity is defined:
        self.u_pts = np.column_stack((x_n,y_n))
        self.u_shape = self.u_X.shape

        self.map_ij2k_u = []

        for i, row in enumerate(self.u_X):
            for j, val in enumerate(row):
                self.map_ij2k_u.append([i,j])

        self.map_ij2k_u = np.asarray(self.map_ij2k_u)

    def makeLaplacian(self, bound = {'N':'value','S':'value','E':'value','W':'value'}):
        """
        Calculate a Laplacian operator matrix suitable for solving a 2D Poisson equation
        on a regular Cartesian grid with square boundaries. Note: the graph must have
        equal spacing in the x and y directions (the same delta in x and y directions).

        """

        size_rows = self.cents_shape[0]
        size_cols = self.cents_shape[1]

        sze = size_rows*size_cols

        A = np.zeros((sze,sze))

        for k, (i,j) in enumerate(self.map_ij2k_cents):

            # if we're not on a main boundary:
            if i != 0 and j != 0 and i != size_rows-1 and j != size_cols-1:

                k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])
                k_in1_j = self.map_ij2k_cents.tolist().index([i-1,j])
                k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])
                k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])

                A[k, k_ip1_j] = 1
                A[k, k_in1_j] = 1
                A[k, k_i_jp1] = 1
                A[k, k_i_jn1] = 1
                A[k,k] = -4


            elif i == 0 and j != 0 and j != size_cols -1: # if on the bottom (South) boundary:

                if bound['S'] == 'flux':

                    k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])
                    k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])
                    k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])

                    A[k, k_ip1_j] = 1
                    A[k, k_i_jp1] = 1
                    A[k, k_i_jn1] = 1

                    A[k,k] = -3

                elif bound['S'] == 'value':

                    k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])

                    A[k,k] = 1
                    # A[k,k_ip1_j] = 1

            elif i == size_rows -1 and j != 0 and j != size_cols -1: # if on the top (North) boundary:

                if bound['N'] == 'flux':

                    k_in1_j = self.map_ij2k_cents.tolist().index([i-1,j])
                    k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])
                    k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])

                    A[k, k_in1_j] = 1
                    A[k, k_i_jp1] = 1
                    A[k, k_i_jn1] = 1

                    A[k,k] = -3

                elif bound['N'] == 'value':

                    k_in1_j = self.map_ij2k_cents.tolist().index([i-1,j])

                    A[k,k] = 1
                    # A[k,k_in1_j] = 1

            elif j == 0 and i != 0 and i != size_rows -1:  # if on the left (West) boundary:

                if bound['W'] == 'flux':

                    k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])
                    k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])
                    k_in1_j = self.map_ij2k_cents.tolist().index([i-1,j])

                    A[k, k_i_jp1] = 1
                    A[k, k_ip1_j] = 1
                    A[k, k_in1_j] = 1

                    A[k,k] = -3

                elif bound['W'] == 'value':

                    k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])

                    A[k,k] = 1
                    # A[k,k_i_jp1] = 1

            elif j == size_cols -1 and i != 0 and i != size_rows -1: # if on the right (East) boundary:

                if bound['E'] == 'flux':

                    k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])
                    k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])
                    k_in1_j = self.map_ij2k_cents.tolist().index([i-1,j])

                    A[k, k_i_jn1] = 1
                    A[k, k_ip1_j] = 1
                    A[k, k_in1_j] = 1

                    A[k,k] = -3

                elif bound['E'] == 'value':

                    k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])

                    A[k,k] = 1
                    # A[k,k_i_jn1] = 1

            # corners:
            elif i == 0 and j == 0: # SW corner

                if bound['S'] == 'flux':

                    k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])
                    k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])

                    A[k, k_i_jp1] = 1
                    A[k, k_ip1_j] = 1
                    A[k,k] = -2

                else:
                    A[k,k] = 1

            elif i == size_rows - 1 and j == 0: # NW corner

                if bound['N'] == 'flux':

                    k_in1_j = self.map_ij2k_cents.tolist().index([i - 1,j])
                    k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])

                    A[k, k_i_jp1] = 1
                    A[k, k_in1_j] = 1
                    A[k,k] = -2

                else:
                    A[k,k] = 1

            elif i == 0 and j == size_cols - 1: # SE corner

                if bound['E'] == 'flux':

                    k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])
                    k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])

                    A[k, k_i_jn1] = 1
                    A[k, k_ip1_j] = 1
                    A[k,k] = -2

                else:
                    A[k,k] = 1

            elif i == size_rows - 1 and j == size_cols - 1: # NE corner

                if bound['E'] == 'flux':

                    k_in1_j = self.map_ij2k_cents.tolist().index([i - 1,j])
                    k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])

                    A[k, k_i_jn1] = 1
                    A[k, k_in1_j] = 1
                    A[k,k] = -2

                else:
                    A[k,k] = 1


        A = A/(self.delta**2)

        # calculate the inverse, which is stored for solution calculation of Laplace and Poisson equations
        Ainv = np.linalg.pinv(A)

        return A, Ainv

    def makeScreenedLaplacian(self, ko=1.0e9, bound = {'N':'value','S':'value','E':'value','W':'value'}):
        """
        Calculate a Laplacian operator matrix suitable for solving a 2D Poisson equation
        on a regular Cartesian grid with square boundaries. Note: the graph must have
        equal spacing in the x and y directions (the same delta in x and y directions).

        """

        size_rows = self.cents_shape[0]
        size_cols = self.cents_shape[1]

        sze = size_rows*size_cols

        A = np.zeros((sze,sze))

        for k, (i,j) in enumerate(self.map_ij2k_cents):

            # if we're not on a main boundary:
            if i != 0 and j != 0 and i != size_rows-1 and j != size_cols-1:

                k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])
                k_in1_j = self.map_ij2k_cents.tolist().index([i-1,j])
                k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])
                k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])

                A[k, k_ip1_j] = 1
                A[k, k_in1_j] = 1
                A[k, k_i_jp1] = 1
                A[k, k_i_jn1] = 1
                A[k,k] = -4 - (self.delta**2)*(ko**2)


            elif i == 0 and j != 0 and j != size_cols -1: # if on the bottom (South) boundary:

                if bound['S'] == 'flux':

                    k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])
                    k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])
                    k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])

                    A[k, k_ip1_j] = 1
                    A[k, k_i_jp1] = 1
                    A[k, k_i_jn1] = 1

                    A[k,k] = -3 - (self.delta**2)*(ko**2)

                elif bound['S'] == 'value':

                    k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])

                    A[k,k] = 1
                    # A[k,k_ip1_j] = 1

            elif i == size_rows -1 and j != 0 and j != size_cols -1: # if on the top (North) boundary:

                if bound['N'] == 'flux':

                    k_in1_j = self.map_ij2k_cents.tolist().index([i-1,j])
                    k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])
                    k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])

                    A[k, k_in1_j] = 1
                    A[k, k_i_jp1] = 1
                    A[k, k_i_jn1] = 1

                    A[k,k] = -3 - (self.delta**2)*(ko**2)

                elif bound['N'] == 'value':

                    k_in1_j = self.map_ij2k_cents.tolist().index([i-1,j])

                    A[k,k] = 1
                    # A[k,k_in1_j] = 1

            elif j == 0 and i != 0 and i != size_rows -1:  # if on the left (West) boundary:

                if bound['W'] == 'flux':

                    k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])
                    k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])
                    k_in1_j = self.map_ij2k_cents.tolist().index([i-1,j])


                    A[k, k_i_jp1] = 1
                    A[k, k_ip1_j] = 1
                    A[k, k_in1_j] = 1

                    A[k,k] = -3 - (self.delta**2)*(ko**2)

                elif bound['W'] == 'value':

                    k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])

                    A[k,k] = 1
                    # A[k,k_i_jp1] = 1

            elif j == size_cols -1 and i != 0 and i != size_rows -1: # if on the right (East) boundary:

                if bound['E'] == 'flux':

                    k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])
                    k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])
                    k_in1_j = self.map_ij2k_cents.tolist().index([i-1,j])

                    A[k, k_i_jn1] = 1
                    A[k, k_ip1_j] = 1
                    A[k, k_in1_j] = 1

                    A[k,k] = -3 - (self.delta**2)*(ko**2)

                elif bound['E'] == 'value':

                    k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])

                    A[k,k] = 1
                    # A[k,k_i_jn1] = 1

            # corners:
            elif i == 0 and j == 0: # SW corner

                if bound['S'] == 'flux':

                    k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])
                    k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])

                    A[k, k_i_jp1] = 1
                    A[k, k_ip1_j] = 1
                    A[k,k] = -2 - (self.delta**2)*(ko**2)

                else:
                    A[k,k] = 1

            elif i == size_rows - 1 and j == 0: # NW corner

                if bound['N'] == 'flux':

                    k_in1_j = self.map_ij2k_cents.tolist().index([i - 1,j])
                    k_i_jp1 = self.map_ij2k_cents.tolist().index([i,j+1])

                    A[k, k_i_jp1] = 1
                    A[k, k_in1_j] = 1
                    A[k,k] = -2 - (self.delta**2)*(ko**2)

                else:
                    A[k,k] = 1

            elif i == 0 and j == size_cols - 1: # SE corner

                if bound['E'] == 'flux':

                    k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1,j])
                    k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])

                    A[k, k_i_jn1] = 1
                    A[k, k_ip1_j] = 1
                    A[k,k] = -2 - (self.delta**2)*(ko**2)

                else:
                    A[k,k] = 1

            elif i == size_rows - 1 and j == size_cols - 1: # NE corner

                if bound['E'] == 'flux':

                    k_in1_j = self.map_ij2k_cents.tolist().index([i - 1,j])
                    k_i_jn1 = self.map_ij2k_cents.tolist().index([i,j-1])

                    A[k, k_i_jn1] = 1
                    A[k, k_in1_j] = 1
                    A[k,k] = -2 - (self.delta**2)*(ko**2)

                else:
                    A[k,k] = 1


        A = A/(self.delta**2)

        # calculate the inverse, which is stored for solution calculation of Laplace and Poisson equations
        Ainv = np.linalg.pinv(A)

        return A, Ainv

    def makeLaplacian_u(self, bound = {'N':'value','S':'value','E':'value','W':'value'}):
        """
        Calculate a Laplacian operator matrix suitable for solving a 2D Poisson equation
        on a regular Cartesian grid with square boundaries. Note: the graph must have
        equal spacing in the x and y directions (the same delta in x and y directions).

        """

        size_rows = self.u_shape[0]
        size_cols = self.u_shape[1]

        sze = size_rows*size_cols

        A = np.zeros((sze,sze))

        for k, (i,j) in enumerate(self.map_ij2k_u):

            # if we're not on a main boundary:
            if i != 0 and j != 0 and i != size_rows-1 and j != size_cols-1:

                k_ip1_j = self.map_ij2k_u.tolist().index([i + 1,j])
                k_in1_j = self.map_ij2k_u.tolist().index([i-1,j])
                k_i_jp1 = self.map_ij2k_u.tolist().index([i,j+1])
                k_i_jn1 = self.map_ij2k_u.tolist().index([i,j-1])

                A[k, k_ip1_j] = 1
                A[k, k_in1_j] = 1
                A[k, k_i_jp1] = 1
                A[k, k_i_jn1] = 1
                A[k,k] = -4


            if i == 0: # if on the bottom (South) boundary:

                if bound['S'] == 'flux':

                    k_ip1_j = self.map_ij2k_u.tolist().index([i + 1,j])
                    A[k, k_ip1_j] = 1

                    A[k,k] = -1

                elif bound['S'] == 'value':

                    k_ip1_j = self.map_ij2k_u.tolist().index([i + 1,j])

                    A[k,k] = 1
                    A[k,k_ip1_j] = 1

            if i == size_rows -1: # if on the top (North) boundary:

                if bound['N'] == 'flux':

                    k_in1_j = self.map_ij2k_u.tolist().index([i-1,j])
                    A[k, k_in1_j] = 1

                    A[k,k] = -1

                elif bound['N'] == 'value':

                    k_in1_j = self.map_ij2k_u.tolist().index([i-1,j])

                    A[k,k] = 1
                    A[k,k_in1_j] = 1

            if j == 0: # if on the left (West) boundary:

                if bound['W'] == 'flux':

                    k_i_jp1 = self.map_ij2k_u.tolist().index([i,j+1])
                    A[k, k_i_jp1] = 1

                    A[k,k] = -1

                elif bound['W'] == 'value':

                    k_i_jp1 = self.map_ij2k_u.tolist().index([i,j+1])

                    A[k,k] = 1
                    A[k,k_i_jp1] = 1

            if j == size_cols -1: # if on the right (East) boundary:

                if bound['E'] == 'flux':

                    k_i_jn1 = self.map_ij2k_u.tolist().index([i,j-1])
                    A[k, k_i_jn1] = 1

                    A[k,k] = -1

                elif bound['E'] == 'value':

                    k_i_jn1 = self.map_ij2k_u.tolist().index([i,j-1])

                    A[k,k] = 1
                    A[k,k_i_jn1] = 1

        A = A/(self.delta**2)

        # calculate the inverse, which is stored for solution calculation of Laplace and Poisson equations
        Ainv = np.linalg.pinv(A)

        return A, Ainv

    def makeLaplacian_v(self, bound = {'N':'value','S':'value','E':'value','W':'value'}):
        """
        Calculate a Laplacian operator matrix suitable for solving a 2D Poisson equation
        on a regular Cartesian grid with square boundaries. Note: the graph must have
        equal spacing in the x and y directions (the same delta in x and y directions).

        """

        size_rows = self.v_shape[0]
        size_cols = self.v_shape[1]

        sze = size_rows*size_cols

        A = np.zeros((sze,sze))

        for k, (i,j) in enumerate(self.map_ij2k_v):

            # if we're not on a main boundary:
            if i != 0 and j != 0 and i != size_rows-1 and j != size_cols-1:

                k_ip1_j = self.map_ij2k_v.tolist().index([i + 1,j])
                k_in1_j = self.map_ij2k_v.tolist().index([i-1,j])
                k_i_jp1 = self.map_ij2k_v.tolist().index([i,j+1])
                k_i_jn1 = self.map_ij2k_v.tolist().index([i,j-1])

                A[k, k_ip1_j] = 1
                A[k, k_in1_j] = 1
                A[k, k_i_jp1] = 1
                A[k, k_i_jn1] = 1
                A[k,k] = -4


            if i == 0: # if on the bottom (South) boundary:

                if bound['S'] == 'flux':

                    k_ip1_j = self.map_ij2k_v.tolist().index([i + 1,j])
                    A[k, k_ip1_j] = 1

                    A[k,k] = -1

                elif bound['S'] == 'value':

                    k_ip1_j = self.map_ij2k_v.tolist().index([i + 1,j])

                    A[k,k] = 1
                    A[k,k_ip1_j] = 1

            if i == size_rows -1: # if on the top (North) boundary:

                if bound['N'] == 'flux':

                    k_in1_j = self.map_ij2k_v.tolist().index([i-1,j])
                    A[k, k_in1_j] = 1

                    A[k,k] = -1

                elif bound['N'] == 'value':

                    k_in1_j = self.map_ij2k_v.tolist().index([i-1,j])

                    A[k,k] = 1
                    A[k,k_in1_j] = 1

            if j == 0: # if on the left (West) boundary:

                if bound['W'] == 'flux':

                    k_i_jp1 = self.map_ij2k_v.tolist().index([i,j+1])
                    A[k, k_i_jp1] = 1

                    A[k,k] = -1

                elif bound['W'] == 'value':

                    k_i_jp1 = self.map_ij2k_v.tolist().index([i,j+1])

                    A[k,k] = 1
                    A[k,k_i_jp1] = 1

            if j == size_cols -1: # if on the right (East) boundary:

                if bound['E'] == 'flux':

                    k_i_jn1 = self.map_ij2k_v.tolist().index([i,j-1])
                    A[k, k_i_jn1] = 1

                    A[k,k] = -1

                elif bound['E'] == 'value':

                    k_i_jn1 = self.map_ij2k_v.tolist().index([i,j-1])

                    A[k,k] = 1
                    A[k,k_i_jn1] = 1

        A = A/(self.delta**2)

        # calculate the inverse, which is stored for solution calculation of Laplace and Poisson equations
        Ainv = np.linalg.pinv(A)

        return A, Ainv

    def makeIntegrator(self):
        """
        Calculate a Finite Difference integration operator matrix for the
        2D grid. Stencil takes 1/2 of central point and 1/8 of points interpolated
        to the side of each imaginary grid square.

        """

        size_rows = self.cents_shape[0]
        size_cols = self.cents_shape[1]

        sze = size_rows*size_cols

        A = np.zeros((sze,sze))

        for k, (i, j) in enumerate(self.map_ij2k_cents):

            # if we're not on a main boundary:
            if i != 0 and j != 0 and i != size_rows - 1 and j != size_cols - 1:

                k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1, j])
                k_in1_j = self.map_ij2k_cents.tolist().index([i - 1, j])
                k_i_jp1 = self.map_ij2k_cents.tolist().index([i, j + 1])
                k_i_jn1 = self.map_ij2k_cents.tolist().index([i, j - 1])

                A[k, k_ip1_j] = 1 / 16
                A[k, k_in1_j] = 1 / 16
                A[k, k_i_jp1] = 1 / 16
                A[k, k_i_jn1] = 1 / 16
                A[k, k] = (1 / 2) + (1 / 4)


            elif i == 0 and j != 0 and j != size_cols - 1:  # if on the bottom (South) boundary:

                k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1, j])
                k_i_jp1 = self.map_ij2k_cents.tolist().index([i, j + 1])
                k_i_jn1 = self.map_ij2k_cents.tolist().index([i, j - 1])

                A[k, k_ip1_j] = 1 / 16
                A[k, k_i_jp1] = 1 / 16
                A[k, k_i_jn1] = 1 / 16

                A[k, k] = (1 / 2) + (3 / 16)


            elif i == size_rows - 1 and j != 0 and j != size_cols - 1:  # if on the top (North) boundary:

                k_in1_j = self.map_ij2k_cents.tolist().index([i - 1, j])
                k_i_jp1 = self.map_ij2k_cents.tolist().index([i, j + 1])
                k_i_jn1 = self.map_ij2k_cents.tolist().index([i, j - 1])

                A[k, k_in1_j] = 1 / 16
                A[k, k_i_jp1] = 1 / 16
                A[k, k_i_jn1] = 1 / 16

                A[k, k] = (1 / 2) + (3 / 16)

            elif j == 0 and i != 0 and i != size_rows - 1:  # if on the left (West) boundary:

                k_i_jp1 = self.map_ij2k_cents.tolist().index([i, j + 1])
                k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1, j])
                k_in1_j = self.map_ij2k_cents.tolist().index([i - 1, j])

                A[k, k_i_jp1] = 1 / 16
                A[k, k_ip1_j] = 1 / 16
                A[k, k_in1_j] = 1 / 16

                A[k, k] = (1 / 2) + (3 / 16)

            elif j == size_cols - 1 and i != 0 and i != size_rows - 1:  # if on the right (East) boundary:

                k_i_jn1 = self.map_ij2k_cents.tolist().index([i, j - 1])
                k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1, j])
                k_in1_j = self.map_ij2k_cents.tolist().index([i - 1, j])

                A[k, k_i_jn1] = 1 / 16
                A[k, k_ip1_j] = 1 / 16
                A[k, k_in1_j] = 1 / 16

                A[k, k] = (1 / 2) + (3 / 16)

            # corners:
            elif i == 0 and j == 0:  # SW corner

                k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1, j])
                k_i_jp1 = self.map_ij2k_cents.tolist().index([i, j + 1])

                A[k, k_i_jp1] = 1 / 16
                A[k, k_ip1_j] = 1 / 16
                A[k, k] = (1 / 2) + (1 / 8)

            elif i == size_rows - 1 and j == 0:  # NW corner

                k_in1_j = self.map_ij2k_cents.tolist().index([i - 1, j])
                k_i_jp1 = self.map_ij2k_cents.tolist().index([i, j + 1])

                A[k, k_i_jp1] = 1 / 16
                A[k, k_in1_j] = 1 / 16
                A[k, k] = A[k, k] = (1 / 2) + (1 / 8)

            elif i == 0 and j == size_cols - 1:  # SE corner

                k_ip1_j = self.map_ij2k_cents.tolist().index([i + 1, j])
                k_i_jn1 = self.map_ij2k_cents.tolist().index([i, j - 1])

                A[k, k_i_jn1] = 1 / 16
                A[k, k_ip1_j] = 1 / 16
                A[k, k] = (1 / 2) + (1 / 8)

            elif i == size_rows - 1 and j == size_cols - 1:  # NE corner

                k_in1_j = self.map_ij2k_cents.tolist().index([i - 1, j])
                k_i_jn1 = self.map_ij2k_cents.tolist().index([i, j - 1])

                A[k, k_i_jn1] = 1 / 16
                A[k, k_in1_j] = 1 / 16
                A[k, k] = (1 / 2) + (1 / 8)

        return A


    #FIXME: This method is currently broken and requires repair. Pampas grass!
    def stokes_kernel(self):
        '''
        Calculate the linearized Navier-Stokes equations using the Finite
        Difference method on a Marker and Cell (MACs) grid.
        '''

        # parameters of the liquid
        rho = 1e3   # density
        visc = 0.1  # visocity

        # body force on the liquid, with components at the x and y coordinates of velocity:
        Fy = np.zeros(self.v_shape)
        Fy[5:-5,5:-5]=-9.81*rho

        Fx = np.zeros(self.u_shape)

        #FIXME: This method returns a tuple of matrices rather than a single
        #matrix. The inverse matrix is the second value of this tuple. Flowers!
        Ainv = self.makeLaplacian()

        time_step = 1e-5
        end_time = 1e-3
        time_points = end_time/time_step
        time = np.linspace(0,end_time,time_points)

        # Unsteady Stokes Flow Solver on MACs grids

        # Initial conditions:

        # Velocity u= ui + vj
        u = np.zeros(self.u_shape)
        v = np.zeros(self.v_shape)

        P_time = []
        u_time = []
        v_time = []

        for t in time:

            # reinforce boundary conditions
            #left
            u[:,0] = 0
            # right
            u[:,-1] = 0
            # top
            u[-1,:] = 0
            # bottom
            u[0,:] = 0

            # left
            v[:,0] = 0
            # right
            v[:,-1] = 0
            # top
            v[-1,:] = 0
            # bottom
            v[0,:] = 0

            # calculate the flow, omitting the pressure term:
            lap_u = laplacian(u,self.delta_u_x,self.delta_u_y)
            lap_v = laplacian(v,self.delta_v_x,self.delta_v_y)

            u = u + (time_step/rho)*(visc*lap_u + Fx)

            v = v + (time_step/rho)*(visc*lap_v + Fy)

            # take the divergence of the flow field using a forward difference
            # that creates a matrix the same size as the pressure matrix:

            u_dx = (u[:,1:] - u[:,0:-1])/self.delta_u_x
            v_dy = (v[1:,:] - v[0:-1,:])/self.delta_v_y

            div_u = u_dx + v_dy

            #...solve for the pressure in terms of existing divergence:

            source = (rho/time_step)*div_u.ravel()
    #         source[bvals] = 0

            P = np.dot(Ainv, source)
            P = P.reshape(self.cents_shape)

            # enforce zero gradient boundary conditions on P:
            P[:,0] = P[:,1]
            P[:,-1] = P[:,-2]
            P[0,:] = P[1,:]
            P[-1,:] = P[-2,:]

            # Take the gradient of the pressue:
            gPxo, gPyo = gradient(P,self.delta_cents_x,self.delta_cents_y)

            gPx = np.zeros(self.u_shape)
            gPx[:,0:-1] = gPxo
            gPx[:,-1] = gPxo[:,-1]

            gPy = np.zeros(self.v_shape)
            gPy[0:-1,:] = gPyo
            gPy[-1,:] = gPyo[-1,:]

            # subtract the pressure from the solution to yeild a divergence-free flow field
            u = u - gPx*(time_step/rho)
            v = v - gPy*(time_step/rho)

            # interpolate u and v values at the centre for easy plotting:
            u_at_c = u[:,0:-1]

            v_at_c = v[0:-1,:]

            # reinforce boundary conditions
            #left
            u_at_c[:,0] = 0
            # right
            u_at_c[:,-1] = 0
            # top
            u_at_c[-1,:] = 0
            # bottom
            u_at_c[0,:] = 0

            # left
            v_at_c[:,0] = 0
            # right
            v_at_c[:,-1] = 0
            # top
            v_at_c[-1,:] = 0
            # bottom
            v_at_c[0,:] = 0

            P_time.append(P)
            u_time.append(u_at_c)
            v_time.append(v_at_c)

    def grid_gradient(self,P,bounds='closed'):
        """
        Calculates a gradient for the MACs grid for a property P
        defined on MACs cell centres.

        Allows boundary conditions to be set: 'closed' means no flux out of boundary
        'open' means an exchange of flux can occur.

        """

        # gradient for ecm:
        gPx = np.zeros(self.u_shape)
        gPy = np.zeros(self.v_shape)

        gPxo = (P[:,1:] - P[:,0:-1])/(2*self.delta)
        gPyo = (P[1:,:] - P[0:-1,:])/(2*self.delta)

        gPx[:,1:-1] = gPxo
        gPy[1:-1,:] = gPyo

        # decide on boundary conditions:
        if bounds == 'open':
            # no "acceleration" on any boundary:
            gPx[:,0] = gPx[:,1]
            gPx[:,-1] = gPx[:,-2]
            gPx[0,:] = gPx[1,:]
            gPx[-1,:] = gPx[-2,:]

            gPy[0,:] = gPy[1,:]
            gPy[-1,:] = gPy[-2,:]
            gPy[:,0] = gPy[:,1]
            gPy[:,-1] = gPy[:,-2]

        elif bounds == 'closed':
            # no flux:
            gPx[:,0] = 0
            gPx[:,-1] = 0
            gPy[0,:] = 0
            gPy[-1,:] = 0

        elif bounds == 'none':
            pass

        return gPx, gPy

    def grid_int(self,F,bounds='closed'):

        # interpolate F to x and y midpoints:
        Fx = np.zeros(self.u_shape)
        Fy = np.zeros(self.v_shape)

        Fxo = (F[:,1:] + F[:,0:-1])/2
        Fyo = (F[1:,:] + F[0:-1,:])/2

        Fx[:,1:-1] = Fxo
        Fy[1:-1,:] = Fyo

        # set boundary conditions:
        if bounds == 'open':
            # zero "gradient" on any boundary:
            Fx[:,0] = Fx[:,1]
            Fx[:,-1] = Fx[:,-2]
            # Fx[0,:] = Fx[1,:]
            # Fx[-1,:] = Fx[-2,:]

            Fy[0,:] = Fy[1,:]
            Fy[-1,:] = Fy[-2,:]
            # Fy[:,0] = Fy[:,1]
            # Fy[:,-1] = Fy[:,-2]

        elif bounds == 'closed':
            # no flux:
            Fx[:,0] = 0
            Fx[:,-1] = 0
            Fy[0,:] = 0
            Fy[-1,:] = 0

        elif bounds == 'none':
            pass

        F_int = np.zeros(F.shape)

        eP = Fx[:,1:] # east midpoints
        wP = Fx[:,0:-1] # west midpoints
        nP = Fy[1:,:] # north midpoints
        sP = Fy[0:-1,:] # south midpoints

        # do the numerical integration:
        F_int[:,:] = (1/2)*F
        F_int[:,:] = F_int[:,:] + (1/8)*nP
        F_int[:,:] = F_int[:,:] + (1/8)*sP
        F_int[:,:] = F_int[:,:] + (1/8)*eP
        F_int[:,:] = F_int[:,:] + (1/8)*wP

        return F_int

def jacobi(A,b,N=50,x=None):
    """
    Solves the equation Ax=b via the Jacobi iterative method.
    Courtesy of Michael Halls-Moore
    https://www.quantstart.com/articles/Jacobi-Method-in-Python-and-NumPy
    """
    # Create an initial guess if needed
    if x is None:
        x = np.zeros(len(A[0]))

    # Create a vector of the diagonal elements of A
    # and subtract them from A
    D = np.diag(A)
    R = A - np.diagflat(D)

    # Iterate for N times
    for i in range(N):
        xo = x
        x = (b - np.dot(R,x)) / D
        er = np.sum((x-xo)**2)/len(x)

    return x, er

def laplacian(F,delx,dely=None):
    """
    Uses discrete central finite difference to calculate the action of a Laplacian operator on a
    dataset F.

    Parameters
    ------------
    F:          A Numpy ndarray defined as F(x,y) on a grid
    dx:         Spacing of the grid (must be uniform in x and y directions)

    Returns
    -----------
    del2F        The Laplacian of F.

    """

    if dely is None:
        dely = delx

    ddFx_interior = (F[:,2:] - 2*F[:,1:-1] + F[:,0:-2])/(delx*dely)
    ddFy_interior = (F[2:,:] - 2*F[1:-1,:] + F[0:-2,:])/(delx*dely)

    ddF_B = (F[-3,:] - 2*F[-2,:] + F[-1,:])/(dely*dely)
    ddF_T = (F[0,:] - 2*F[-1,:] + F[-2,:])/(dely*dely)
    ddF_L = (F[:,2] - 2*F[:,1] + F[:,0])/(delx*delx)
    ddF_R = (F[:,-3] - 2*F[:,-2] + F[:,-1])/(delx*delx)

    lapF_interior = ddFx_interior[1:-1,:] + ddFy_interior[:,1:-1]

    # initialize the dFx and dFy arrays:
    ddF = np.zeros(F.shape)

    ddF[1:-1,1:-1] = lapF_interior

    ddF[:,0] = ddF_L
    ddF[:,-1] = ddF_R

    ddF[0,:] = ddF_T
    ddF[-1,:] = ddF_B


    return ddF

def gradient(F,delx,dely=None):
    # gradient using numpy slicing:

    if dely is None:
        dely = delx

    # calculate the discrete central first derivatives on the internal mesh points:
    dF_interior_y = -(F[:-2,:] - F[2:,:])/(2*dely)
    dF_interior_x = -(F[:,:-2] - F[:,2:])/(2*delx)

    # calculate the discrete forward or backward first derivatives on the boundary points:
    dF_B = (F[1,:] - F[0,:])/dely
    dF_T = (F[-1,:] - F[-2,:])/dely
    dF_L = (F[:,1] - F[:,0])/delx
    dF_R = (F[:,-1] - F[:,-2])/delx

    # initialize the dFx and dFy arrays:
    dFx = np.zeros(F.shape)
    dFy = np.zeros(F.shape)

    # build the final dFx and dFy arrays by splicing together internal and boundary derivatives:
    dFx[:,1:-1] = dF_interior_x
    dFy[1:-1,:] = dF_interior_y

    dFx[:,0] = dF_L
    dFx[:,-1] = dF_R

    dFy[0,:] = dF_B
    dFy[-1,:] = dF_T

    return dFx, dFy

def diff(F,delx,axis=0):
    # dertivative using numpy slicing:

    if axis == 1:
        # calculate the discrete central first derivatives on the internal mesh points:
        dF_interior = -(F[:-2,:] - F[2:,:])/(2*delx)

        # calculate the discrete forward or backward first derivatives on the boundary points:
        dF_B = -(F[1,:] - F[0,:])/delx
        dF_T = -(F[-1,:] - F[-2,:])/delx

        dF = np.zeros(F.shape)

        dF[1:-1,:] = dF_interior

        dF[0,:] = dF_B
        dF[-1,:] = dF_T


    elif axis == 0:
        # calculate the discrete central first derivatives on the internal mesh points:
        dF_interior = -(F[:,:-2] - F[:,2:])/(2*delx)

        # calculate the discrete forward or backward first derivatives on the boundary points:
        dF_L = (F[:,0] - F[:,1])/delx
        dF_R = (F[:,-2] - F[:,-1])/delx

        dF = np.zeros(F.shape)

        dF[:,1:-1] = dF_interior

        dF[:,0] = dF_L
        dF[:,-1] = dF_R


    return dF

def divergence(Fx,Fy,delx,dely):

    gx = diff(Fx, delx, axis=0)
    gy = diff(Fy, dely, axis=1)
    div = gx + gy

    return div

def curl(Fx, Fy, delx,dely):

    """
    Calculates the curl of a 2D vector field.
    The returned parameter is the z-axis component (in or out of the page).

    """

    gyx = diff(Fy,delx,axis=0)
    gxy = diff(Fx,dely,axis=1)

    curlF = gyx - gxy

    return curlF

def makeMask(M,xy_pts,X,Y,delta,sensitivity=1.0):
    """
    Creates an nd-array matrix mask with in-boundary
    elements tagged as 1, on-boundary elements tagged
    as -1 and environmental points tagged as 0.

    Parameters
    -----------
    M               A matrix with non-zero data and zero environment
    xy_pts          unravelled xy points for the Cells of M
    X               2D X matrix for the Cells of M
    Y               2D Y matrix for the Cells of M
    delta           Spacing (average x and y) for world of M grid
    sensitivity     1.0 is average, lower selects less boundary points, higher more

    Returns
    -----------
    maskM           The mask matrix
    """

    maskM = np.zeros(M.shape)
    inds = (M > 0).nonzero()
    maskM[inds] = 1

    x_bpts = X[inds].ravel()
    y_bpts =Y[inds].ravel()
    bpts = np.column_stack((x_bpts,y_bpts))

    bound_inds = boundTag(bpts,delta,alpha=sensitivity)
    points_tree = cKDTree(xy_pts)
    b_inds = points_tree.query(bpts[bound_inds])[1]

    mm = maskM.ravel()
    mm[b_inds] = -1
    maskM = mm.reshape(M.shape)

    return maskM

def boundTag(points,delta,alpha=1.0):
    """
    Flag elements that are on the boundary to the environment by calculating
    the convex hull for a points cluster.

    Parameters
    ----------
    points          A numpy array of [x,y] points. This may be ecm_verts_unique, cell_centres, or mem_mids_flat.


    Returns
    -------
    bflags       A python list of indices of points that are on the boundary
    bmask        A numpy array of boolean flags of points that are on the boundary (order indexed to points_Flat)

    Notes
    -------
    Uses numpy arrays
    Uses alpha_shape function to calculate the concave hull
    Requires a nested input such as self.mem_mids or self.ecm_verts

    """
    alpha = alpha/delta

    tri = Delaunay(points)
    tri_edges = []
    # circum_r_list = []

    for ia, ib, ic in tri.vertices:
        pa = points[ia]
        pb = points[ib]
        pc = points[ic]

        # Lengths of sides of triangle
        a = math.sqrt((pa[0]-pb[0])**2 + (pa[1]-pb[1])**2)
        b = math.sqrt((pb[0]-pc[0])**2 + (pb[1]-pc[1])**2)
        c = math.sqrt((pc[0]-pa[0])**2 + (pc[1]-pa[1])**2)

        # Semiperimeter of triangle
        s = (a + b + c)/2.0

        # Area of triangle by Heron's formula
        area = math.sqrt(abs(s*(s-a)*(s-b)*(s-c)))

        if area > 0:
            circum_r = a*b*c/(4.0*area)

        if area == 0:
            circum_r = a*b*c/(4.0*1e-25)

        # Here's the radius filter:
        if circum_r < 1.0/alpha:
            #self.tri_edges_append([ia, ib])
            tri_edges.append([ia, ib])
            tri_edges.append([ib, ic])
            tri_edges.append([ia, ic])

    for i, edge in enumerate(tri_edges):  # First organize the list so that all [i,j] and [j,i] are equalized
        pt1 = edge[0]
        pt2 = edge[1]
        if pt1 > pt2:
            tri_edges[i]=[pt2,pt1]

    tri_edges.sort()

    tri_edges_len = len(tri_edges)
    concave_hull = []

    i = 0
    # Now step through to find and remove all duplicating entities and add them to a new list
    while i < tri_edges_len:
        j = i + 1
        tri_edge_i = tri_edges[i]

        while j < tri_edges_len and tri_edge_i == tri_edges[j]:
            j += 1

        if i == j - 1:
            concave_hull.append(tri_edge_i)
            i += 1
        else:
            i = j

#     con_hull = tb.alpha_shape(points, alpha/delta)  # get the concave hull for the membrane midpoints
    concave_hull = np.asarray(concave_hull)

    bflags = np.unique(concave_hull)    # get the value of unique indices from segments

    return bflags

# def integrator(P):
#
#     F = np.zeros(P.shape)
#
#     eP = P[:,1:] # east midpoints
#     wP = P[:,0:-1] # west midpoints
#     nP = P[1:,:] # north midpoints
#     sP = P[0:-1,:] # south midpoints
#
#     F[:, :] = (1 / 2) * P
#     F[0:-1, :] += (1 / 8) * nP
#     F[1:, :] += (1 / 8) * sP
#     F[:, 0:-1] += (1 / 8) * eP
#     F[:, 1:] += (1 / 8) * wP
#
#     # reset boundary values:
#     F[:, 0] = P[:, 0]
#     F[:, -1] = P[:, -1]
#     F[0, :] = P[0, :]
#     F[-1, :] = P[-1, :]
#
#     return F

def integrator(P, sharp = 0.5):
    """
    Averages nearest neighbours of the environmental array with a weighting
    given by the "sharp" option.

    P: some 2D matrix
    sharp: weighting of the neigbouring averages; 0.5 is standard finite volume smoothing; 1.0 is no smoothing

    Thanks Sess!

    """

    F = np.zeros(P.shape)

    eP = P[:,1:] # east midpoints
    wP = P[:,0:-1] # west midpoints
    nP = P[1:,:] # north midpoints
    sP = P[0:-1,:] # south midpoints

    sides = (1-sharp)/4

    F[:, :] = sharp * P
    F[0:-1, :] += sides * nP
    F[1:, :] += sides * sP
    F[:, 0:-1] += sides * eP
    F[:, 1:] += sides * wP

    # reset boundary values:
    F[:, 0] = P[:, 0]
    F[:, -1] = P[:, -1]
    F[0, :] = P[0, :]
    F[-1, :] = P[-1, :]

    return F

# flux summer
def flux_summer(U,V,P):

    F = np.zeros(P.shape)

    eU = U[:,1::1] # east midpoints
    wU = U[:,0:-1:1] # west midpoints
    nV = V[1::1,:] # north midpoints
    sV = V[0:-1:1,:] # south midpoints

    # do the sum:
    F[:,:] = nV - sV + eU - wU

    return F

def flux_summer_sym(U,V):

    F = np.zeros(U.shape)

    eP = U[:,1:] # east midpoints
    wP = U[:,0:-1] # west midpoints
    nP = V[1:,:] # north midpoints
    sP = V[0:-1,:] # south midpoints

    F[0:-1,:] = nP
    F[1:,:] = F[1:,:] - sP
    F[:,0:-1] = F[:,0:-1] + eP
    F[:,1:] = F[:,1:] - wP

    return F





#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **venv** (i.e., lightweight virtual environment containing a
distinct Python binary and site directories, optionally isolated from system
site directories) facilities.
'''

# ....................{ IMPORTS                           }....................
import sys
# from betse.util.io.log import logs
from betse.util.type.decorator.decmemo import func_cached
# from betse.util.type.types import type_check

# ....................{ TESTERS                           }....................
#FIXME: Generalize this tester to detect Anaconda-specific venvs as well. See:
#    https://stackoverflow.com/a/40099080/2809027
@func_cached
def is_venv() -> bool:
    '''
    ``True`` only if the active Python interpreter is isolated to a venv.

    Specifically, this function returns ``True`` only if this interpreter is
    isolated to a venv produced by either:

    * The official :mod:`venv` package bundled with Python >= 3.3.
    * The third-party :mod:`virtualenv` package supporting both Python 2 and 3.

    See Also
    ----------
    https://stackoverflow.com/a/42580137/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # Return true if this interpreter is isolated to a venv produced by
    # either...
    return (
        # "virtualenv", which uniquely defines the "sys.real_prefix"
        # attribute to the absolute dirname of the top-level directory
        # containing the system-wide Python interpreter *OR*...
        hasattr(sys, 'real_prefix') or

        # "venv", which (possibly non-uniquely) sets:
        #
        # * The "sys.base_prefix" attribute to the absolute dirname of the
        #   top-level directory containing the system-wide Python interpreter.
        # * The "sys.prefix" attribute to the absolute dirname of the
        #   top-level directory containing the venv-specific Python interpreter
        #   if any *OR* the system-wide Python interpreter otherwise.
        #
        # Note that, as Python >= 3.3 *ALWAYS* defines the "sys.base_prefix"
        # attribute, testing this attribute's existence is unnecessary.
        sys.prefix != sys.base_prefix
    )

# ....................{ GETTERS                           }....................
#FIXME: Generalize this tester to support Anaconda-specific venvs as well.
@func_cached
def get_system_prefix() -> bool:
    '''
    Absolute dirname of the top-level directory containing the system-wide
    Python interpreter, regardless of whether the active Python interpreter is
    isolated to a venv.
    '''

    # Return either...
    return (
        # The "virtualenv"-specific attribute providing this dirname if the
        # the active Python interpreter is isolated to a venv *OR*...
        sys.real_prefix
        if hasattr(sys, 'real_prefix') else
        # The standard attribute providing this dirname otherwise.
        sys.base_prefix
    )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising *ML (e.g., HTML, SGML, XML) functionality.
'''

# ....................{ IMPORTS                           }....................
# import pytest

# ....................{ TESTS                             }....................
def test_is_ml() -> None:
    '''
    Unit test the :func:`betse.util.type.text.mls.is_ml` tester.
    '''

    # Defer heavyweight imports.
    from betse.util.type.text import mls

    # Assert this tester to behave as expected.
    assert mls.is_ml(
        '<nausicaa kaze no="tane">And that one shall come to you</nausicaa>')
    assert not mls.is_ml(
        '<garbed in "raiment" of blue and descending upon a field of gold>...')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests for BETSE's Python API testing the importability of fragile public
modules and packages (e.g., containing cyclic import statements subject to
breaking by brute-force PyCharm refactorings).
'''

# ....................{ IMPORTS                           }....................
from betse.util.test.pytest.mark.pytfail import xfail
# from betse_test.mark.skip import skip_unless_plugin_xdist

# ....................{ TESTS                             }....................
#FIXME: Ugh! Ideally, this test should *NOT* require xdist support. This test
#must be isolated to a subprocess. To do so, the simplest means is the xdist
#"--boxed" CLI option. However, we'd only want *THIS* test to be isolated in
#that manner -- not *EVERY* test! To do so, we'll probably want to examine the
#xdist codebase, ascertain how "--boxed" is implemented, and provide a new
#utility test function or decorator performing the equivalent. Alternately,
#PyInstaller's testing framework already supports test isolation (probably using
#the "subprocess" module), suggesting we could also examine that as a fallback.

# @skip_unless_plugin_xdist
@xfail(reason='Test isolation unsupported.')
def test_import_logs() -> None:
    '''
    Test the importability of BETSE's logging API.

    This API contains cyclic import statements commonly broken by PyCharm's
    brute-force refactorings.
    '''

    # Import this API and call all functionality containing these statements.
    from betse.util.io.log import logs
    logs.get_logger()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level :mod:`pytest` fixtures.
'''

# ....................{ IMPORTS                            }....................
from pytest import fixture

# ....................{ FIXTURES                           }....................
@fixture(scope='session')
def monkeypatch_session():
    '''
    Session-scoped ``monkeypatch`` fixture.

    Caveats
    -------
    **The ``monkeypatch`` fixture should never be referenced from session-scoped
    fixtures.** Doing so raises the following fatal error on session startup:

        ScopeMismatch: You tried to access the 'function' scoped fixture
        'monkeypatch' with a 'session' scoped request object, involved
        factories

    This equivalent fixture suffers no such deficits and hence should be
    referenced from session-scoped fixtures instead.

    Returns
    -------
    _pytest.monkeypatch.MonkeyPatch
        Object returned by the `monkeypatch` fixture keeping a record of
        :func:`setattr`, item, environment, and :attr:`sys.path` changes.

    See Also
    --------
    https://github.com/pytest-dev/pytest/issues/363#issuecomment-406536200
        GitHub comment strongly inspiring this implementation.
    '''

    # Defer violations of privacy encapsulation for safety. Tragically, this
    # class is inaccessible via any other means.
    from _pytest.monkeypatch import MonkeyPatch

    # Create and yield to the calling session-scoped fixture a new
    # monkey-patcher instance.
    monkey_patch = MonkeyPatch()
    yield monkey_patch

    # Revert all temporary changes applied to this instance by the calling
    # session-scoped fixture.
    monkey_patch.undo()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# ....................{ IMPORTS ~ fixture                 }....................
'''
Global functional test configuration for BETSE's command-line interface (CLI).

``py.test`` implicitly imports all functionality defined by this module into
all CLI-specific functional test modules. Since this functionality now includes
all publicly declared functional fixtures in this ``fixture`` subpackage, these
tests may reference these fixtures without explicit imports.
'''

# ....................{ IMPORTS ~ fixture                 }....................
from betse_test.a90_func._fixture.clier import betse_cli



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level standard error facilities.
'''

# ....................{ IMPORTS                           }....................
import sys, traceback
from betse.util.type.types import type_check, StrOrNoneTypes

# ....................{ OUTPUTTERS                        }....................
@type_check
def output(*texts: str) -> None:
    '''
    Print all passed strings to stderr *without* logging these strings.

    This function is intentionally *not* named :func:`print`, as doing so would
    invite conflicts with the standard :func:`print` function.
    '''

    print(*texts, file=sys.stderr)


@type_check
def output_warning(*warnings: str) -> None:
    '''
    Print all passed strings (prefixed by a suitable human-readable warning
    label) to stderr *without* logging these strings.
    '''

    # ...that was easier than expected.
    output('WARNING: ', *warnings)

# ....................{ OUTPUTTERS ~ exception            }....................
@type_check
def output_exception(heading: StrOrNoneTypes = None) -> None:
    '''
    Print the currently caught exception to stderr *without* logging this
    exception optionally preceded by the passed human-readable heading if any.

    Parameters
    ----------
    heading : optional[str]
        Optional human-readable heading to be printed before this exception if
        any *or* ``None`` if no heading is to be printed.
    '''

    #FIXME: Assert that an exception has actually been raised here.

    # If a heanding is passed, print this heading to stderr.
    if heading is not None:
        output(heading)

    # Print this exception to stderr.
    traceback.print_exc(file=sys.stderr)


def output_traceback() -> None:
    '''
    Print the current call stack to stderr *without* logging this call stack.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.call import callers

    # Print this call stack, excluding the calls to both this and the
    # callers.get_traceback() functions.
    output(callers.get_traceback(-2))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests for the :mod:`betse.lib.yaml` subpackage.
'''

# ....................{ IMPORTS                            }....................
from betse_test._fixture.simconf.simconfclser import SimConfTestInternal

# ....................{ TESTS                              }....................
def test_yaml_copy(
    betse_sim_conf: SimConfTestInternal,
    betse_temp_dir: 'py._path.local.LocalPath',
) -> None:
    '''
    Test the capacity of the :mod:`betse.lib.yaml.yamls` submodule to reliably
    copy a deserialized simulation configuration (including both the top-level
    YAML file for this configuration *and* all external paths internally
    referenced and hence required by this file).

    Parameters
    ----------
    betse_sim_conf : SimConfTestInternal
        Object encapsulating a temporary simulation configuration file.
    betse_temp_dir : py._path.local.LocalPath
        Object encapsulating a temporary directory isolated to this test.
    '''

    # Absolute dirname of a subdirectory with arbitrary basename in this
    # temporary directory.
    new_subdirpath = betse_temp_dir.join('Ten_Forward')

    # Create this subdirectory
    new_subdirpath.ensure(dir=True)

    # Absolute filename of a new simulation configuration file with arbitrary
    # basename to be created in this subdirectory.
    new_sim_conf_filepath = new_subdirpath.join('El-Aurian_Guinan.yaml')

    # Absolute filename of this file as a string rather than a "LocalPath"
    # instance.
    new_sim_conf_filename = str(new_sim_conf_filepath)

    # Simulation configuration loaded from this file.
    p = betse_sim_conf.p

    # Copy this configuration to this subdirectory.
    p.save(conf_filename=new_sim_conf_filename)

    # Assert this file to have been created.
    assert new_sim_conf_filepath.check(file=1)


def test_yaml_roundtrip(betse_sim_conf: SimConfTestInternal) -> None:
    '''
    Test the capacity of the :mod:`betse.lib.yaml.yamls` submodule to reliably
    roundtrip (i.e., load and save without loss) the default YAML-formatted
    simulation configuration file, whose complexity warrants explicit testing.

    Parameters
    ----------
    betse_sim_conf : SimConfTestInternal
        Object encapsulating a temporary simulation configuration file.
    '''

    # Simulation configuration loaded from this file.
    p = betse_sim_conf.p

    # Absolute filename of this file.
    p_conf_filename = p.conf_filename

    #FIXME: Test addition, deletion, and modification of non-scalar list items.

    # Modify an arbitrary boolean setting of this configuration.
    p.is_ecm = not p.is_ecm
    p_sim_ECM_expected = p.is_ecm

    # Modify an arbitrary numeric setting of this configuration.
    p.cell_polarizability *= 3.1415
    p_cell_polarizability_expected = p.cell_polarizability

    # Modify an arbitrary string setting of this configuration.
    p.seed_pickle_basename += '~'
    p_seed_pickle_basename_expected = p.seed_pickle_basename

    # Save these changes back to the same file.
    p.save_inplace()

    # Close this file.
    p.unload()

    # Read these changes back into the same in-memory object.
    p.load(p_conf_filename)

    # Ensure these changes were roundtripped across this I/O.
    assert p.is_ecm == p_sim_ECM_expected
    assert p.cell_polarizability == p_cell_polarizability_expected
    assert p.seed_pickle_basename == p_seed_pickle_basename_expected



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Error haiku** (i.e., pseudo-random human-readable general-purpose error
message formatted as a haiku) facilities.
'''

# ....................{ IMPORTS                           }....................
import random, sys, traceback

# ....................{ CONSTANTS                         }....................
ERROR_HAIKU = (
    (
        'Chaos reigns within.',
        'Reflect, repent, and reboot.',
        'Order shall return.',
    ), (
        'ABORTED effort:',
        'Close all that you have.',
        'You ask way too much.',
    ), (
        'First snow, then silence.',
        'This thousand dollar screen dies',
        'so beautifully.',
    ), (
        'A crash reduces',
        'your expensive computer',
        'to a simple stone.',
    ), (
        'Error messages',
        'cannot completely convey.',
        'We now know shared loss.',
    ), (
        'The code was willing.',
        'It considered your request',
        'but the chips were weak.',
    ), (
        'There is a chasm',
        'of carbon and silicon',
        "this software can't bridge.",
    ), (
        'To have no errors',
        'would be life without meaning.',
        'No struggle, no joy.',
    ), (
        'many fingers clicking',
        'screens are full of letters',
        'what is their meaning?',
    ), (
        'Water spills downwards.',
        'Electric stream cascades sparks.',
        'Data flow ceases.',
    ), (
        'Technical support',
        'would be a flowing source of',
        'sweet commiseration.',
    ), (
        'Fatal exception.',
        'Code has looped upon itself',
        'like the coiled serpent.',
    ), (
        'An old CPU.',
        'A new program is loaded.',
        'The sound of crashing.',
    ),
)
'''
List of haikus to be printed in the event of fatal errors.

This is serious business, folks.

See Also
----------
http://baetzler.de/humor/haiku_error.var
https://www.gnu.org/fun/jokes/error-haiku.html
    To quote: "IMAGINE IF INSTEAD OF CRYPTIC TEXT STRINGS, YOUR COMPUTER
    PRODUCED ERROR MESSAGES IN HAIKU..." We need no longer imagine.
'''

# ....................{ GETTERS                           }....................
def get_random() -> str:
    '''
    Pseudo-random human-readable general-purpose error message formatted as a
    haiku, commonly printed and/or logged in the event of fatal errors.
    '''

    return '\n'.join(random.choice(ERROR_HAIKU))



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level operating system (OS) facilities.

Caveats
----------
**Operating system-specific logic is poor form.** Do so *only* where necessary.
'''

# ....................{ IMPORTS                           }....................
import platform
from betse import metadata
from betse.exceptions import BetseOSException
from betse.util.io.log import logs
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict

# ....................{ INITIALIZERS                      }....................
def init() -> None:
    '''
    Validate the current platform.

    This function (in order):

    #. Logs a non-fatal warning if this platform is a non-WSL variant
       of Microsoft Windows (e.g., vanilla Windows, Cygwin Windows).
    #. Logs a non-fatal warning if this platform is *not* recognized as
       officially supported by this application (e.g., BSD*, Solaris).
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import windows

    # Log this validation.
    logs.log_debug('Validating platform...')

    # Human-readable string describing the set of all officially supported
    # platforms known to interoperate sanely with this application.
    oses_supported_str = (
        'Consider running {name} only under Linux or macOS. '
        'Note that Linux is now freely emulatable under Windows 10 '
        'via the Windows Subsystem for Linux (WSL). '
        'See official installation instructions at:\n'
        '\thttps://docs.microsoft.com/en-us/windows/wsl/install-win10'.format(
            name=metadata.NAME))

    # If this is a non-WSL Windows variant, log a non-fatal warning.
    if windows.is_windows():
        #FIXME: Restore this logging statement to a non-fatal warning *AFTER*
        #resolving the "FIXME:" comment chain in the
        #"betse.util.os.brand.windows" submodule concerning PowerShell. Yikes!
        # logs.log_warning(
        logs.log_info(
            'Windows platform detected. '
            'Python itself and third-party scientific frameworks for Python '
            '(e.g., Numpy, SciPy, Matplotlib) are known to '
            'behave suboptimally on this platform. '
            '%s',
            oses_supported_str)

    # If this platform is officially unsupported by this application, log a
    # non-fatal warning.
    if not is_supported():
        logs.log_warning(
            'Unsupported platform "%s" detected. %s',
            get_name(), oses_supported_str)

# ....................{ EXCEPTIONS                        }....................
def die_if_unsupported() -> bool:
    '''
    Raise an exception if the current platform is officially unsupported by
    this application.

    Raises
    ------
    BetseOSException
        If this platform is officially unsupported by this application.

    See Also
    ----------
    :func:`is_supported`
        Further details.
    '''

    # If this platform is unsupported, raise an exception.
    if not is_supported():
        raise BetseOSException('Platform "{}" unsupported'.format(get_name()))

# ....................{ TESTERS                           }....................
@func_cached
def is_supported() -> bool:
    '''
    ``True`` only if the current platform is officially supported by
    this application.

    This function currently only returns ``True`` for the following platforms
    (in no particular order):

    * Apple macOS.
    * Linux.
    * Microsoft Windows.

    Caveats
    ----------
    This function returning ``True`` does *not* necessarily imply this
    application to behave optimally under this platform. In particular,
    Microsoft Windows is officially supported by popular demand but well-known
    to behave suboptimally with respect to Python itself and third-party
    scientific frameworks for Python (e.g., Numpy, Matplotlib).
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import linux, macos, windows

    # The drawing of the Three draws nigh.
    return linux.is_linux() or macos.is_macos() or windows.is_windows()

# ....................{ GETTERS                           }....................
@func_cached
def get_name() -> str:
    '''
    Human-readable name of the current platform.

    This function returns:

    * Under Linux, the name of this Linux distribution, detected as follows:

      * If the :func:`platform.linux_distribution` function is available, the
        first element of the 3-tuple returned by this function (e.g.,
        ``CentOS``).
      * Else, the string returned by the :func:`platform.system` function.
        Since this is probably the generic string ``Linux``, this is only a
        fallback.

    * Under macOS, ``macOS``.
    * Under Windows, the flavour of the operating system-level kernel exposed
      by the shell environment to which the active Python interpreter is
      attached. Specifically, if this is:

      * The Windows Subsystem for Linux (WSL), ``Windows (WSL)``. Note,
        however, that the WSL masquerades as Linux. Thankfully, it does so to a
        remarkably more reliable degree than Cygwin Windows.
      * Cygwin Windows, ``Windows (Cygwin)``. Note, however, that Cygwin
        masquerades as Linux, replicating the POSIX process model and GNU/Linux
        userland to a remarkable degree -- subject to the Big List of Dodgy
        Apps (BLODA), proprietary limitations of the Win32 API, and similar
        caveats.
      * Under vanilla Windows shell environments (e.g., CMD.exe, PowerShell),
        merely ``Windows``.

    * Under all other platforms, the string returned by the
      :func:`platform.system` function.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import linux, macos, windows

    # Platform name to be returned.
    os_name = None

    # If Linux, defer to the "linux" submodule.
    if linux.is_linux():
        os_name = linux.get_distro_name()
    # If macOS, return "macOS". Since platform.system() returns the low-level
    # kernel name "Darwin", this name is ignored.
    elif macos.is_macos():
        os_name = 'macOS'
    # If Cygwin Windows, return "Windows (Cygwin)". Since platform.system()
    # returns a non-human-readable low-level uppercase label specific to the
    # word size of the current Python interpreter (e.g., "CYGWIN_NT-5.1*"),
    # this label is ignored.
    elif windows.is_windows_cygwin():
        os_name = 'Windows (Cygwin)'
    # If non-Cygwin Windows, return merely "Windows". The name returned by
    # platform.system() depends on the current Windows version as follows:
    #
    # * If this is Windows Vista, this name is "Microsoft".
    # * Else, this name is "Windows".
    #
    # Hence, this name is ignored.
    elif windows.is_windows_vanilla():
        os_name = 'Windows'
    # Else, reuse the name returned by the prior call to platform.system().

    # If no platform-specific name was detected above, default to the platform
    # name returned by the platform.system() function. Since this typically
    # corresponds to the low-level name of the current kernel (e.g., "Darwin",
    # "Linux") rather than the high-level name of the current platform (e.g.,
    # "macOS", "CentOS"), this is only a fallback of last resort.
    if os_name is None:
        os_name = platform.system()

    # Return the name established above.
    return os_name


@func_cached
def get_version() -> str:
    '''
    Human-readable ``.``-delimited version specifier string of the current
    platform.

    This function returns:

    * Under Linux, the version of the current Linux distribution reported by
      the :func:`betse.util.os.brand.linux.get_distro_version` function.
    * Under macOS, this installation's current major and minor version (e.g.,
      ``10.9``).
    * Under Windows, this installation's current major version (e.g., ``8``).
      Unfortunately, there appears to be no consistently reliable means of
      obtaining this system's current major *and* minor version (e.g.,
      ``8.1``).
    * Under all other platforms, the string returned by the
      :func:`platform.release` function.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import linux, macos

    # Version specifier to be returned.
    os_version = None

    # If Linux, defer to the "linux" submodule.
    if linux.is_linux():
        os_version = linux.get_distro_version()
    # If macOS *AND* the platform.mac_ver() function is available, return the
    # first item of the 3-tuple returned by this function (e.g., "10.9"). Since
    # platform.release() returns the low-level kernel version (e.g., "13.0.0"),
    # this version is ignored when feasible.
    elif macos.is_macos() and hasattr(platform, 'mac_ver'):
        # platform.mac_ver() returns a 3-tuple (release, versioninfo, machine)
        # of macOS-specific metadata, where "versioninfo" is itself a 3-tuple
        # (version, dev_stage, non_release_version). Return the high-level
        # "release" element (e.g., "10.9") rather than the optionally defined
        # low-level "version" element of the "versioninfo" element, which is
        # typically the empty string and hence useless.
        os_version = platform.mac_ver()[0]
    # If Windows, accept the default version specifier returned by
    # platform.release() (e.g., "8", "10"). Since this specifier appears to
    # coincide exactly with the first element of the 4-tuple returned by the
    # conditionally available platform.win32_ver() function, there is no
    # benefit to the latter approach.

    # If no platform-specific version specifier was detected above, default to
    # the version specifier returned by the platform.release() function. Since
    # this typically corresponds to the low-level version of the current kernel
    # (e.g., "4.1.15") rather than the high-level version of the current OS
    # (e.g., "6.4"), this is merely a fallback of last resort.
    if os_version is None:
        os_version = platform.release()

    # Return this version specifier.
    return os_version

# ....................{ GETTERS ~ metadata                }....................
def get_metadata() -> OrderedArgsDict:
    '''
    Ordered dictionary synopsizing the current platform.
    '''

    # Return this dictionary.
    return OrderedArgsDict(
        'name', get_name(),
        'version', get_version(),
    )
